"""
This file contains helper functions for building the model and for loading model parameters.
These helper functions are built to mirror those in the official TensorFlow implementation.
"""

import re
import math
import collections
from functools import partial
import torch
from torch import nn
from torch.nn import functional as F
from torch.utils import model_zoo
from .utils_extra import Conv2dStaticSamePadding

########################################################################
############### HELPERS FUNCTIONS FOR MODEL ARCHITECTURE ###############
########################################################################


# Parameters for the entire model (stem, all blocks, and head)

GlobalParams = collections.namedtuple('GlobalParams', [
    'batch_norm_momentum', 'batch_norm_epsilon', 'dropout_rate',
    'num_classes', 'width_coefficient', 'depth_coefficient',
    'depth_divisor', 'min_depth', 'drop_connect_rate', 'image_size'])

# Parameters for an individual model block
BlockArgs = collections.namedtuple('BlockArgs', [
    'kernel_size', 'num_repeat', 'input_filters', 'output_filters',
    'expand_ratio', 'id_skip', 'stride', 'se_ratio'])

# Change namedtuple defaults
GlobalParams.__new__.__defaults__ = (None,) * len(GlobalParams._fields)
BlockArgs.__new__.__defaults__ = (None,) * len(BlockArgs._fields)


class SwishImplementation(torch.autograd.Function):
    @staticmethod
    def forward(ctx, i):
        result = i * torch.sigmoid(i)
        ctx.save_for_backward(i)
        return result

    @staticmethod
    def backward(ctx, grad_output):
        i = ctx.saved_variables[0]
        sigmoid_i = torch.sigmoid(i)
        return grad_output * (sigmoid_i * (1 + i * (1 - sigmoid_i)))


class MemoryEfficientSwish(nn.Module):
    def forward(self, x):
        return SwishImplementation.apply(x)


class Swish(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(x)


def round_filters(filters, global_params):
    """ Calculate and round number of filters based on depth multiplier. """
    multiplier = global_params.width_coefficient
    if not multiplier:
        return filters
    divisor = global_params.depth_divisor
    min_depth = global_params.min_depth
    filters *= multiplier
    min_depth = min_depth or divisor
    new_filters = max(min_depth, int(filters + divisor / 2) // divisor * divisor)
    if new_filters < 0.9 * filters:  # prevent rounding by more than 10%
        new_filters += divisor
    return int(new_filters)


def round_repeats(repeats, global_params):
    """ Round number of filters based on depth multiplier. """
    multiplier = global_params.depth_coefficient
    if not multiplier:
        return repeats
    return int(math.ceil(multiplier * repeats))


def drop_connect(inputs, p, training):
    """ Drop connect. """
    if not training: return inputs
    batch_size = inputs.shape[0]
    keep_prob = 1 - p
    random_tensor = keep_prob
    random_tensor += torch.rand([batch_size, 1, 1, 1], dtype=inputs.dtype, device=inputs.device)
    binary_tensor = torch.floor(random_tensor)
    output = inputs / keep_prob * binary_tensor
    return output


def get_same_padding_conv2d(image_size=None):
    """ Chooses static padding if you have specified an image size, and dynamic padding otherwise.
        Static padding is necessary for ONNX exporting of models. """
    if image_size is None:
        return Conv2dDynamicSamePadding
    else:
        return partial(Conv2dStaticSamePadding, image_size=image_size)


class Conv2dDynamicSamePadding(nn.Conv2d):
    """ 2D Convolutions like TensorFlow, for a dynamic image size """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, dilation=1, groups=1, bias=True):
        super().__init__(in_channels, out_channels, kernel_size, stride, 0, dilation, groups, bias)
        self.stride = self.stride if len(self.stride) == 2 else [self.stride[0]] * 2

    def forward(self, x):
        ih, iw = x.size()[-2:]
        kh, kw = self.weight.size()[-2:]
        sh, sw = self.stride
        oh, ow = math.ceil(ih / sh), math.ceil(iw / sw)
        pad_h = max((oh - 1) * self.stride[0] + (kh - 1) * self.dilation[0] + 1 - ih, 0)
        pad_w = max((ow - 1) * self.stride[1] + (kw - 1) * self.dilation[1] + 1 - iw, 0)
        if pad_h > 0 or pad_w > 0:
            x = F.pad(x, [pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2])
        return F.conv2d(x, self.weight, self.bias, self.stride, self.padding, self.dilation, self.groups)


class Identity(nn.Module):
    def __init__(self, ):
        super(Identity, self).__init__()

    def forward(self, input):
        return input


########################################################################
############## HELPERS FUNCTIONS FOR LOADING MODEL PARAMS ##############
########################################################################


def efficientnet_params(model_name):
    """ Map EfficientNet model name to parameter coefficients. """
    params_dict = {
        # Coefficients:   width,depth,res,dropout
        'efficientnet-b0': (1.0, 1.0, 224, 0.2),
        'efficientnet-b1': (1.0, 1.1, 240, 0.2),
        'efficientnet-b2': (1.1, 1.2, 260, 0.3),
        'efficientnet-b3': (1.2, 1.4, 300, 0.3),
        'efficientnet-b4': (1.4, 1.8, 380, 0.4),
        'efficientnet-b5': (1.6, 2.2, 456, 0.4),
        'efficientnet-b6': (1.8, 2.6, 528, 0.5),
        'efficientnet-b7': (2.0, 3.1, 600, 0.5),
        'efficientnet-b8': (2.2, 3.6, 672, 0.5),
        'efficientnet-l2': (4.3, 5.3, 800, 0.5),
    }
    return params_dict[model_name]


class BlockDecoder(object):
    """ Block Decoder for readability, straight from the official TensorFlow repository """

    @staticmethod
    def _decode_block_string(block_string):
        """ Gets a block through a string notation of arguments. """
        assert isinstance(block_string, str)

        ops = block_string.split('_')
        options = {}
        for op in ops:
            splits = re.split(r'(\d.*)', op)
            if len(splits) >= 2:
                key, value = splits[:2]
                options[key] = value

        # Check stride
        assert (('s' in options and len(options['s']) == 1) or
                (len(options['s']) == 2 and options['s'][0] == options['s'][1]))

        return BlockArgs(
            kernel_size=int(options['k']),
            num_repeat=int(options['r']),
            input_filters=int(options['i']),
            output_filters=int(options['o']),
            expand_ratio=int(options['e']),
            id_skip=('noskip' not in block_string),
            se_ratio=float(options['se']) if 'se' in options else None,
            stride=[int(options['s'][0])])

    @staticmethod
    def _encode_block_string(block):
        """Encodes a block to a string."""
        args = [
            'r%d' % block.num_repeat,
            'k%d' % block.kernel_size,
            's%d%d' % (block.strides[0], block.strides[1]),
            'e%s' % block.expand_ratio,
            'i%d' % block.input_filters,
            'o%d' % block.output_filters
        ]
        if 0 < block.se_ratio <= 1:
            args.append('se%s' % block.se_ratio)
        if block.id_skip is False:
            args.append('noskip')
        return '_'.join(args)

    @staticmethod
    def decode(string_list):
        """
        Decodes a list of string notations to specify blocks inside the network.

        :param string_list: a list of strings, each string is a notation of block
        :return: a list of BlockArgs namedtuples of block args
        """
        assert isinstance(string_list, list)
        blocks_args = []
        for block_string in string_list:
            blocks_args.append(BlockDecoder._decode_block_string(block_string))
        return blocks_args

    @staticmethod
    def encode(blocks_args):
        """
        Encodes a list of BlockArgs to a list of strings.

        :param blocks_args: a list of BlockArgs namedtuples of block args
        :return: a list of strings, each string is a notation of block
        """
        block_strings = []
        for block in blocks_args:
            block_strings.append(BlockDecoder._encode_block_string(block))
        return block_strings


def efficientnet(width_coefficient=None, depth_coefficient=None, dropout_rate=0.2,
                 drop_connect_rate=0.2, image_size=None, num_classes=1000):
    """ Creates a efficientnet model. """

    blocks_args = [
        'r1_k3_s11_e1_i32_o16_se0.25', 'r2_k3_s22_e6_i16_o24_se0.25',
        'r2_k5_s22_e6_i24_o40_se0.25', 'r3_k3_s22_e6_i40_o80_se0.25',
        'r3_k5_s11_e6_i80_o112_se0.25', 'r4_k5_s22_e6_i112_o192_se0.25',
        'r1_k3_s11_e6_i192_o320_se0.25',
    ]
    blocks_args = BlockDecoder.decode(blocks_args)

    global_params = GlobalParams(
        batch_norm_momentum=0.99,
        batch_norm_epsilon=1e-3,
        dropout_rate=dropout_rate,
        drop_connect_rate=drop_connect_rate,
        # data_format='channels_last',  # removed, this is always true in PyTorch
        num_classes=num_classes,
        width_coefficient=width_coefficient,
        depth_coefficient=depth_coefficient,
        depth_divisor=8,
        min_depth=None,
        image_size=image_size,
    )

    return blocks_args, global_params


def get_model_params(model_name, override_params):
    """ Get the block args and global params for a given model """
    if model_name.startswith('efficientnet'):
        w, d, s, p = efficientnet_params(model_name)
        # note: all models have drop connect rate = 0.2
        blocks_args, global_params = efficientnet(
            width_coefficient=w, depth_coefficient=d, dropout_rate=p, image_size=s)
    else:
        raise NotImplementedError('model name is not pre-defined: %s' % model_name)
    if override_params:
        # ValueError will be raised here if override_params has fields not included in global_params.
        global_params = global_params._replace(**override_params)
    return blocks_args, global_params


url_map = {
    'efficientnet-b0': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b0-355c32eb.pth',
    'efficientnet-b1': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b1-f1951068.pth',
    'efficientnet-b2': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b2-8bb594d6.pth',
    'efficientnet-b3': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b3-5fb5a3c3.pth',
    'efficientnet-b4': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b4-6ed6700e.pth',
    'efficientnet-b5': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b5-b6417697.pth',
    'efficientnet-b6': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b6-c76e70fd.pth',
    'efficientnet-b7': 'https://publicmodels.blob.core.windows.net/container/aa/efficientnet-b7-dcc49843.pth',
}

url_map_advprop = {
    'efficientnet-b0': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b0-b64d5a18.pth',
    'efficientnet-b1': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b1-0f3ce85a.pth',
    'efficientnet-b2': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b2-6e9d97e5.pth',
    'efficientnet-b3': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b3-cdd7c0f4.pth',
    'efficientnet-b4': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b4-44fb3a87.pth',
    'efficientnet-b5': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b5-86493f6b.pth',
    'efficientnet-b6': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b6-ac80338e.pth',
    'efficientnet-b7': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b7-4652b6dd.pth',
    'efficientnet-b8': 'https://publicmodels.blob.core.windows.net/container/advprop/efficientnet-b8-22a8fe65.pth',
}


def load_pretrained_weights(model, model_name, load_fc=True, advprop=False):
    """ Loads pretrained weights, and downloads if loading for the first time. """
    # AutoAugment or Advprop (different preprocessing)
    url_map_ = url_map_advprop if advprop else url_map
    state_dict = model_zoo.load_url(url_map_[model_name], map_location=torch.device('cpu'))
    # state_dict = torch.load('../../weights/backbone_efficientnetb0.pth')
    if load_fc:
        ret = model.load_state_dict(state_dict, strict=False)
        print(ret)
    else:
        state_dict.pop('_fc.weight')
        state_dict.pop('_fc.bias')
        res = model.load_state_dict(state_dict, strict=False)
        assert set(res.missing_keys) == set(['_fc.weight', '_fc.bias']), 'issue loading pretrained weights'
    print('Loaded pretrained weights for {}'.format(model_name))



from .resnet_backbone import resnet18
from torch import nn
import torch
import torch.nn.functional as F
from detro.networks.components import BiFPN, Center_layer, Offset_layer, Reg_layer, Heatmap_layer
from detro.networks.losslib import center_loss, distance_loss


class FeatureFusionNetwork(nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, inputs):
        resized = []
        size = inputs[0].size()[-2:]
        for x in inputs[1:]:
            resized.append(F.upsample(x, size))
        x = torch.cat(resized, dim=1)
        return x




class CircleNet(nn.Module):
    def __init__(self, num_classes=1):
        super().__init__()
        self.backbone = resnet18(pretrained=True)
        self.neck = FeatureFusionNetwork()
        self.conv1 = nn.Conv2d(896, 256, kernel_size=1, stride=1, padding=0)
        self.bn1 = nn.BatchNorm2d(256)
        self.relu = nn.ReLU(inplace=True)
        # self.center_layer = Heatmap_layer(in_channels=256, out_channels=num_classes)
        # self.reg_layer = Heatmap_layer(in_channels=256, out_channels=1)
        self.hm_layer = Heatmap_layer(in_channels=256, out_channels=num_classes + 1)

    def forward(self, inputs):
        c1, c2, c3, c4, c5 = self.backbone(inputs)
        features = [c2, c3, c4, c5]
        features = self.neck(features)
        x = features
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        # center_heatmap = self.center_layer(x)
        # offsets = self.reg_layer(x)
        x=self.hm_layer(x)
        center_heatmap=x[:,:-1]
        offsets=x[:,-1:]
        return dict(
            center_heatmap=center_heatmap, offsets=offsets
        )


def CircleDetCriterion(preds, labels):
    loss_center = center_loss(preds['center_heatmap'], labels['center_heatmap'])
    # loss_corner=center_loss(preds['corner_heatmap'],labels['corner_heatmap'])
    loss_offsets = distance_loss(preds['offsets'], labels['offsets'], labels['offsets_mask'])
    return dict(
        loss=loss_center + loss_offsets,
        loss_center=loss_center,
        # loss_corner=loss_corner,
        loss_offsets=loss_offsets,
    )



from torch import nn


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=dilation, groups=groups, bias=False, dilation=dilation)


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)
class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(BasicBlock, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None, groups=1,
                 base_width=64, dilation=1, norm_layer=None):
        super(Bottleneck, self).__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


import math
import numpy as np


def _get_rotate_info(angle, cx, cy, h, w):
    import math
    import numpy as np
    angle = math.radians(angle)
    cos = math.cos(angle)
    sin = math.sin(angle)
    M = np.array([[cos, sin], [-sin, cos]])
    cos = abs(cos)
    sin = abs(sin)
    nw = w * cos + h * sin
    nh = w * sin + h * cos
    return M, nw, nh


def rotate_points(points, angle, cx, cy, h, w):
    import numpy as np
    M, nw, nh = _get_rotate_info(angle, cx, cy, h, w)
    points = np.array(points)
    original_shape = points.shape
    points = points.reshape((-1, 2))
    points = M.dot((points - np.array([cx, cy])).T).T + np.array([nw // 2, nh // 2])
    points = points.reshape(original_shape)
    return points


def polar2cartesian(points, from_degree=False):
    points = np.array(points)
    shape = points.shape
    points = points.reshape((-1, 2))
    batch_angle = points[:, 1:2]
    if from_degree:
        batch_angle = np.radians(batch_angle)
    batch_angle -= np.pi
    batch_x = points[:, 0:1] * np.cos(batch_angle)
    batch_y = points[:, 0:1] * np.sin(batch_angle)
    points = np.concatenate((batch_x, batch_y), axis=1)
    points = points.reshape(shape)
    return points


def cartesian2polar(points, to_degree=False):
    points = np.array(points)
    shape = points.shape
    points = points.reshape((-1, 2))
    batch_radius = np.expand_dims(np.sqrt(np.square(points).sum(axis=1)), axis=-1)
    batch_angle = np.expand_dims(np.arctan2(points[:, 1], points[:, 0]), axis=-1) + np.pi
    if to_degree:
        batch_angle = np.degrees(batch_angle)
    points = np.concatenate((batch_radius, batch_angle), axis=1)
    points = points.reshape(shape)
    return points


def bounding_rect(points):
    points = np.array(points)
    l = np.min(points[:, 0])
    r = np.max(points[:, 0])
    t = np.min(points[:, 1])
    b = np.max(points[:, 1])
    return np.array([l, t, r, b])


def pad_quad(quad, pad_size=5, pad_ratio=None, limits=None):
    import math
    from math import inf
    limits = limits or (-inf, -inf, inf, inf)

    def dist(p1, p2):
        return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)

    def mean(li):
        return sum(li) / len(li)

    def minus(la, lb):
        return [x - y for x, y in zip(la, lb)]

    def add(la, lb):
        return [x + y for x, y in zip(la, lb)]

    def get_pad_params(pad_size, pad_ratio, quad):
        # l, t, r, b = box
        p0, p1, p2, p3 = quad
        bh = mean([dist(p0, p3), dist(p1, p2)])
        bw = mean([dist(p0, p1), dist(p3, p2)])
        if pad_ratio is not None:
            if isinstance(pad_ratio, (tuple, list)):
                if len(pad_ratio) == 2:
                    pad_l = pad_r = int(min(bh, bw) * pad_ratio[0])
                    pad_t = pad_b = int(min(bh, bw) * pad_ratio[1])
                else:
                    assert len(pad_ratio) == 4
                    pad_l = int(min(bh, bw) * pad_ratio[0])
                    pad_t = int(min(bh, bw) * pad_ratio[1])
                    pad_r = int(min(bh, bw) * pad_ratio[2])
                    pad_b = int(min(bh, bw) * pad_ratio[3])
            else:
                pad_l = pad_t = pad_r = pad_b = int(min(bh, bw) * pad_ratio)
        else:
            if isinstance(pad_size, (tuple, list)):
                if len(pad_size) == 2:
                    pad_l = pad_r = pad_size[0]
                    pad_t = pad_b = pad_size[1]
                else:
                    assert len(pad_size) == 4
                    pad_l, pad_t, pad_r, pad_b = pad_size
            else:
                pad_l = pad_t = pad_r = pad_b = pad_size
        return pad_l, pad_t, pad_r, pad_b

    pad_l, pad_t, pad_r, pad_b = get_pad_params(pad_size, pad_ratio, quad)
    p0, p1, p2, p3 = quad
    p0 = add(p0, [-pad_l, -pad_t])
    p1 = add(p1, [pad_r, -pad_t])
    p2 = add(p2, [pad_r, pad_b])
    p3 = add(p3, [-pad_l, pad_b])

    return limit_quad([p0, p1, p2, p3], limits=limits)

def points_distance(p0,p1):
    return np.sqrt(np.sum(np.square(np.array(p0)-np.array(p1))))
def points_center(points):
    return np.array(points).mean(axis=0)
def limit_point(p, limits):
    if limits is None: return p
    if len(limits) == 2:
        ml, mt = 0, 0
        mr, mb = limits
    else:
        assert len(limits) == 4
        ml, mt, mr, mb = limits
    x, y = p
    x = max(ml, min(mr, x))
    y = max(mt, min(mb, y))
    return [x, y]


def limit_quad(quad, limits=None):
    quad = [limit_point(p, limits) for p in quad]
    return quad


def organize_quad_points(points):
    points = np.array(points)
    center = points.mean(axis=0)
    angles = np.arctan2(center[1] - points[:, 1], points[:, 0] - center[0]) * 180 / np.pi
    ids = np.argsort(angles)[::-1]
    polygon = points[ids]
    return polygon


def organize_polygon_points(points):
    points = np.array(points)
    center = points.mean(axis=0)
    angles = np.arctan2(center[1] - points[:, 1], points[:, 0] - center[0]) * 180 / np.pi
    ids = np.argsort(angles)[::-1]
    polygon = points[ids]
    return polygon


def polygon_length(points):
    assert len(points) >= 3
    points=list(points)
    points+=[points[0]]
    dist=0
    for i in range(len(points)-1):
        p1=points[i]
        p2=points[i+1]
        dist+=np.sqrt(np.sum(np.square(np.array(p1)-np.array(p2))))
    return dist


def calc_polygon_area(points):
    area = 0
    q = points[-1]
    for p in points:
        area += p[0] * q[1] - p[1] * q[0]
        q = p
    return abs(area / 2)


def calc_quad_angle(quad):
    '''cal angle of a quad'''
    p0, p1, p2, p3 = quad

    def center(p1, p2):
        x1, y1 = p1
        x2, y2 = p2
        return (x1 + x2) / 2, (y1 + y2) / 2

    x1, y1 = center(p0, p3)
    x2, y2 = center(p1, p2)
    angle = math.atan2(y1 - y2, x2 - x1) * 180 / math.pi
    return angle



import torch
import numpy as np
import os
import time

def load_weights(model,weights_path=None,except_keys=[]):
	if weights_path and os.path.exists(weights_path):
		try:
			state_dict=torch.load(weights_path)
			remove_keys=[]
			for k in state_dict.keys():
				for key in except_keys:
					if k.startswith(key):
						remove_keys.append(k)
			for k in remove_keys:
				del state_dict[k]
				print('ignore key:',k)
			model.load_state_dict(state_dict, strict=False)
		except:
			print('Cannot load pretrained model %s' % (weights_path))
			raise
	return model
def train(cfg=None):
	t = time.time()
	time_stamp = time.strftime('%m%d-%H', time.localtime(t))
	model = cfg.get_model()
	cfg.WEIGHTS_INIT = 'weights/training/model.pkl'
	model=load_weights(model,cfg.WEIGHTS_INIT,except_keys=['reg_layer'])
	device = cfg.DEVICE
	model.to(device)
	train_data = cfg.dataset
	optimizer = torch.optim.Adam(model.parameters(), cfg.LEARN_RATE_INIT)
	lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, cfg.MAX_PERIODS, cfg.LEARN_RATE_END)
	min_loss = 1e+10
	for epoch in range(cfg.MAX_PERIODS):
		model.train()
		loss_history = {}
		loss_avg_dict = {}
		loss_summary = []
		for batch_image, batch_labels in train_data:
			preds = model(batch_image)
			loss_dict = cfg.criterion(preds, batch_labels)
			loss = loss_dict['loss']
			optimizer.zero_grad()
			loss.backward()
			optimizer.step()
			for k, v in loss_dict.items():
				if not k in loss_history.keys():
					loss_history[k] = []
				loss_history[k].append(v.item())
		lr_scheduler.step()
		for k, v in loss_history.items():
			loss_avg_dict[k] = np.array(v).mean()
			loss_summary.append('%s:%.4f' % (k, loss_avg_dict[k]))
		loss_epoch = loss_avg_dict['loss']
		loss_summary = '\t'.join(loss_summary)
		print('Epoch: %s\tLearningRate:%.6f\t%s' % (
			epoch, optimizer.state_dict()['param_groups'][0]['lr'], loss_summary))
		if epoch and cfg.WEIGHTS_SAVE_INTERVAL and epoch % cfg.WEIGHTS_SAVE_INTERVAL == 0:
			torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_{time_stamp}_{epoch}_loss_{loss_epoch:.4f}.pkl')
		if loss_epoch < min_loss:
			min_loss = loss_epoch
			torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_best_{time_stamp}.pkl')
			torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_best.pkl')
		torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model.pkl')

if __name__ == '__main__':
	train()



import numpy as np
import math


def normalize(image):
	mean = [0.485, 0.456, 0.406]
	std = [0.229, 0.224, 0.225]
	mean = np.array([[mean]])
	std = np.array([[std]])
	image = (image.astype(np.float32) - mean) / std
	return image


def denormalize(image):
	mean = [0.485, 0.456, 0.406]
	std = [0.229, 0.224, 0.225]
	mean = np.array([[mean]])
	std = np.array([[std]])
	image = (image.astype(np.float32) * std) + mean
	return image


def dist_point_line(p, line):
	x, y = p
	((x0, y0), (x1, y1)) = line

	if x0 == x1:
		return abs(x - x0)
	else:
		line = np.polyfit([x0, x1], [y0, y1], 1)
		k, b = line
		return abs(k * x - y + b) / math.sqrt(k ** 2 + 1)


def line_intersection(line1, line2):
	xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
	ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

	def det(a, b):
		return a[0] * b[1] - a[1] * b[0]

	div = det(xdiff, ydiff)
	if div == 0:
		raise Exception('lines do not intersect')

	d = (det(*line1), det(*line2))
	x = det(d, xdiff) / div
	y = det(d, ydiff) / div
	return x, y


def min_max_vertical_dist(points, line):
	k, b = line
	vdists = points[:, 1] - (k * points[:, 0] + b)
	return min(vdists), max(vdists)


def kb2two_point(line):
	k, b = line
	p1, p2 = [0, b], [10, k * 10 + b]
	return p1, p2


def get_split_line_of_two_quad(quad1, quad2):
	p0, p1 = quad1[1:3]
	p2, p3 = quad2[0], quad2[3]
	p02 = np.mean([p0, p2], axis=0)
	p13 = np.mean([p1, p3], axis=0)
	return [p02, p13]


def approx_polygon(quads):
	q1 = quads[0]
	q2 = quads[-1]
	# p0=np.mean([q1[0],q1[3]],axis=0)
	# p1=np.mean([q2[1],q2[2]],axis=0)

	points = np.array(quads).reshape((-1, 2))
	line = np.polyfit(points[:, 0], points[:, 1], 1)
	k, b = line
	dmin, dmax = min_max_vertical_dist(points, line)
	line0 = kb2two_point([k, b + dmax])
	line2 = kb2two_point([k, b + dmin])
	line3 = [q1[0], q1[3]]
	line1 = [q2[1], q2[2]]

	p0 = line_intersection(line0, line3)
	p1 = line_intersection(line0, line1)
	p2 = line_intersection(line1, line2)
	p3 = line_intersection(line2, line3)
	big_quad = np.array([p0, p1, p2, p3])
	split_points = []
	for i in range(len(quads) - 1):
		l_quad, r_quad = quads[i], quads[i + 1]
		line = get_split_line_of_two_quad(l_quad, r_quad)
		p0 = line_intersection(line, line0)
		p1 = line_intersection(line, line2)
		split_points.append([p0, p1])
	# center=big_quad.mean(axis=0)
	# split_offsets=np.zeros((17*3))
	# quad_offsets=np.zeros()

	res = dict(
		split_points=split_points,
		big_quad=big_quad,
	)
	return res



import wpcv
import detro.utils as ut
import cv2
import os, shutil, glob
import numpy as np

file_exts=['.jpg','.jpeg','.png','.bmp']

def test(cfg):
    ut.set_default_font_path(cfg.FONT_PATH)
    detector = cfg.get_detector()
    num_params=cfg.NUM_REG_PARAMS
    det_type = getattr(cfg,'DET_TYPE','bbox')
    dir=cfg.TEST_DIR
    saver = wpcv.ImageSaver(dir + '_out', remake_dir=True, auto_make_subdir=True)

    fs = []
    for ext in file_exts:
        fs+=glob.glob(dir + '/*'+ext)
    fs.sort()
    font = ut.get_default_font(32)
    for i, f in enumerate(fs):
        img = cv2.imread(f)
        pim = wpcv.pilimg(img)
        polygons = detector.predict(img)
        if not len(polygons):
            continue
        polys = polygons[:, :num_params].astype(np.int)
        # polys=list(filter(lambda poly:poly[num_params]>0.1,))
        im = pim
        for j, poly in enumerate(polys):
            score = polygons[j][num_params]
            label = '%.2f' % score
            if det_type == 'bbox':
                im = wpcv.draw_boxes_with_label(im, [(poly, label)], line_width=2, box_color='red', font=font,
                                                text_color='blue', offset=(0, -18))
            elif det_type=='circle':
               pass
            elif False:
                poly = np.array(poly).reshape((-1, 2))
                box=wpcv.bounding_rect(poly)
                im = wpcv.draw_boxes_with_label(im, [(box, label)], line_width=2, box_color='red', font=font,
                                                text_color='blue', offset=(0, -18))
            else:
                assert det_type == 'polygon'
                poly = np.array(poly).reshape((-1, 2))
                poly = ut.organize_polygon_points(poly)
                im = wpcv.draw_polygon(im, poly, color='blue', width=4, label=label, label_xy=(0, -18),
                                       label_color='blue', font=font)
                # im = wpcv.draw_polygon(im, poly, color='red', width=4, label=label, label_xy=(0, -18),
                #                        label_color='red', font=font)
                if hasattr(detector,'get_middle_results'):
                    res=detector.get_middle_results()
                    saver.save(res['center_heatmap_img'])
                    saver.save(res['corner_heatmap_img'])

        saver.save(im)
        # im=visualize(pim,polygons)
        # saver.save(im)
        print(i, f, len(polygons))




import torch
from torch import nn
from torch.nn import functional as F

from .utils import (
    round_filters,
    round_repeats,
    drop_connect,
    get_same_padding_conv2d,
    get_model_params,
    efficientnet_params,
    load_pretrained_weights,
    Swish,
    MemoryEfficientSwish,
)

class MBConvBlock(nn.Module):
    """
    Mobile Inverted Residual Bottleneck Block

    Args:
        block_args (namedtuple): BlockArgs, see above
        global_params (namedtuple): GlobalParam, see above

    Attributes:
        has_se (bool): Whether the block contains a Squeeze and Excitation layer.
    """

    def __init__(self, block_args, global_params):
        super().__init__()
        self._block_args = block_args
        self._bn_mom = 1 - global_params.batch_norm_momentum
        self._bn_eps = global_params.batch_norm_epsilon
        self.has_se = (self._block_args.se_ratio is not None) and (0 < self._block_args.se_ratio <= 1)
        self.id_skip = block_args.id_skip  # skip connection and drop connect

        # Get static or dynamic convolution depending on image size
        Conv2d = get_same_padding_conv2d(image_size=global_params.image_size)

        # Expansion phase
        inp = self._block_args.input_filters  # number of input channels
        oup = self._block_args.input_filters * self._block_args.expand_ratio  # number of output channels
        if self._block_args.expand_ratio != 1:
            self._expand_conv = Conv2d(in_channels=inp, out_channels=oup, kernel_size=1, bias=False)
            self._bn0 = nn.BatchNorm2d(num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)

        # Depthwise convolution phase
        k = self._block_args.kernel_size
        s = self._block_args.stride
        self._depthwise_conv = Conv2d(
            in_channels=oup, out_channels=oup, groups=oup,  # groups makes it depthwise
            kernel_size=k, stride=s, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=oup, momentum=self._bn_mom, eps=self._bn_eps)

        # Squeeze and Excitation layer, if desired
        if self.has_se:
            num_squeezed_channels = max(1, int(self._block_args.input_filters * self._block_args.se_ratio))
            self._se_reduce = Conv2d(in_channels=oup, out_channels=num_squeezed_channels, kernel_size=1)
            self._se_expand = Conv2d(in_channels=num_squeezed_channels, out_channels=oup, kernel_size=1)

        # Output phase
        final_oup = self._block_args.output_filters
        self._project_conv = Conv2d(in_channels=oup, out_channels=final_oup, kernel_size=1, bias=False)
        self._bn2 = nn.BatchNorm2d(num_features=final_oup, momentum=self._bn_mom, eps=self._bn_eps)
        self._swish = MemoryEfficientSwish()

    def forward(self, inputs, drop_connect_rate=None):
        """
        :param inputs: input tensor
        :param drop_connect_rate: drop connect rate (float, between 0 and 1)
        :return: output of block
        """

        # Expansion and Depthwise Convolution
        x = inputs
        if self._block_args.expand_ratio != 1:
            x = self._expand_conv(inputs)
            x = self._bn0(x)
            x = self._swish(x)

        x = self._depthwise_conv(x)
        x = self._bn1(x)
        x = self._swish(x)

        # Squeeze and Excitation
        if self.has_se:
            x_squeezed = F.adaptive_avg_pool2d(x, 1)
            x_squeezed = self._se_reduce(x_squeezed)
            x_squeezed = self._swish(x_squeezed)
            x_squeezed = self._se_expand(x_squeezed)
            x = torch.sigmoid(x_squeezed) * x

        x = self._project_conv(x)
        x = self._bn2(x)

        # Skip connection and drop connect
        input_filters, output_filters = self._block_args.input_filters, self._block_args.output_filters
        if self.id_skip and self._block_args.stride == 1 and input_filters == output_filters:
            if drop_connect_rate:
                x = drop_connect(x, p=drop_connect_rate, training=self.training)
            x = x + inputs  # skip connection
        return x

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export)"""
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()


class EfficientNet(nn.Module):
    """
    An EfficientNet model. Most easily loaded with the .from_name or .from_pretrained methods

    Args:
        blocks_args (list): A list of BlockArgs to construct blocks
        global_params (namedtuple): A set of GlobalParams shared between blocks

    Example:
        model = EfficientNet.from_pretrained('efficientnet-b0')

    """

    def __init__(self, blocks_args=None, global_params=None):
        super().__init__()
        assert isinstance(blocks_args, list), 'blocks_args should be a list'
        assert len(blocks_args) > 0, 'block args must be greater than 0'
        self._global_params = global_params
        self._blocks_args = blocks_args

        # Get static or dynamic convolution depending on image size
        Conv2d = get_same_padding_conv2d(image_size=global_params.image_size)

        # Batch norm parameters
        bn_mom = 1 - self._global_params.batch_norm_momentum
        bn_eps = self._global_params.batch_norm_epsilon

        # Stem
        in_channels = 3  # rgb
        out_channels = round_filters(32, self._global_params)  # number of output channels
        self._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        self._bn0 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Build blocks
        self._blocks = nn.ModuleList([])
        for block_args in self._blocks_args:

            # Update block input and output filters based on depth multiplier.
            block_args = block_args._replace(
                input_filters=round_filters(block_args.input_filters, self._global_params),
                output_filters=round_filters(block_args.output_filters, self._global_params),
                num_repeat=round_repeats(block_args.num_repeat, self._global_params)
            )

            # The first block needs to take care of stride and filter size increase.
            self._blocks.append(MBConvBlock(block_args, self._global_params))
            if block_args.num_repeat > 1:
                block_args = block_args._replace(input_filters=block_args.output_filters, stride=1)
            for _ in range(block_args.num_repeat - 1):
                self._blocks.append(MBConvBlock(block_args, self._global_params))

        # Head
        in_channels = block_args.output_filters  # output of final block
        out_channels = round_filters(1280, self._global_params)
        self._conv_head = Conv2d(in_channels, out_channels, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm2d(num_features=out_channels, momentum=bn_mom, eps=bn_eps)

        # Final linear layer
        self._avg_pooling = nn.AdaptiveAvgPool2d(1)
        self._dropout = nn.Dropout(self._global_params.dropout_rate)
        self._fc = nn.Linear(out_channels, self._global_params.num_classes)
        self._swish = MemoryEfficientSwish()

    def set_swish(self, memory_efficient=True):
        """Sets swish function as memory efficient (for training) or standard (for export)"""
        self._swish = MemoryEfficientSwish() if memory_efficient else Swish()
        for block in self._blocks:
            block.set_swish(memory_efficient)


    def extract_features(self, inputs):
        """ Returns output of the final convolution layer """

        # Stem
        x = self._swish(self._bn0(self._conv_stem(inputs)))

        # Blocks
        for idx, block in enumerate(self._blocks):
            drop_connect_rate = self._global_params.drop_connect_rate
            if drop_connect_rate:
                drop_connect_rate *= float(idx) / len(self._blocks)
            x = block(x, drop_connect_rate=drop_connect_rate)
        # Head
        x = self._swish(self._bn1(self._conv_head(x)))

        return x

    def forward(self, inputs):
        """ Calls extract_features to extract features, applies final linear layer, and returns logits. """
        bs = inputs.size(0)
        # Convolution layers
        x = self.extract_features(inputs)

        # Pooling and final linear layer
        x = self._avg_pooling(x)
        x = x.view(bs, -1)
        x = self._dropout(x)
        x = self._fc(x)
        return x

    @classmethod
    def from_name(cls, model_name, override_params=None):
        cls._check_model_name_is_valid(model_name)
        blocks_args, global_params = get_model_params(model_name, override_params)
        return cls(blocks_args, global_params)

    @classmethod
    def from_pretrained(cls, model_name, load_weights=True, advprop=True, num_classes=1000, in_channels=3):
        model = cls.from_name(model_name, override_params={'num_classes': num_classes})
        if load_weights:
            load_pretrained_weights(model, model_name, load_fc=(num_classes == 1000), advprop=advprop)
        if in_channels != 3:
            Conv2d = get_same_padding_conv2d(image_size = model._global_params.image_size)
            out_channels = round_filters(32, model._global_params)
            model._conv_stem = Conv2d(in_channels, out_channels, kernel_size=3, stride=2, bias=False)
        return model

    @classmethod
    def get_image_size(cls, model_name):
        cls._check_model_name_is_valid(model_name)
        _, _, res, _ = efficientnet_params(model_name)
        return res

    @classmethod
    def _check_model_name_is_valid(cls, model_name):
        """ Validates model name. """
        valid_models = ['efficientnet-b'+str(i) for i in range(9)]
        if model_name not in valid_models:
            raise ValueError('model_name should be one of: ' + ', '.join(valid_models))



import torch
import torch.nn.functional as F
from torchvision import transforms
import os,shutil,glob
import cv2
# import wpcv
from PIL import Image
import numpy as np
def _nms(heat, kernel=3):
  hmax = F.max_pool2d(heat, kernel, stride=1, padding=(kernel - 1) // 2)
  keep = (hmax == heat).float()
  return heat * keep

def _gather_feature(feat, ind, mask=None):
  dim = feat.size(2)
  ind = ind.unsqueeze(2).expand(ind.size(0), ind.size(1), dim)
  feat = feat.gather(1, ind)
  if mask is not None:
    mask = mask.unsqueeze(2).expand_as(feat)
    feat = feat[mask]
    feat = feat.view(-1, dim)
  return feat

def _topk(scores, K=4):
    batch, height, width = scores.size()

    topk_scores, topk_inds = torch.topk(scores.view(batch, 1, -1), K)

    topk_inds = topk_inds % (height * width)
    topk_ys = (topk_inds / width).int().float()
    topk_xs = (topk_inds % width).int().float()

    topk_score, topk_ind = torch.topk(topk_scores.view(batch, -1), K)
    topk_clses = (topk_ind / K).int()
    topk_inds = _gather_feature(topk_inds.view(batch, -1, 1), topk_ind).view(batch, K)
    topk_ys = _gather_feature(topk_ys.view(batch, -1, 1), topk_ind).view(batch, K)
    topk_xs = _gather_feature(topk_xs.view(batch, -1, 1), topk_ind).view(batch, K)

    return topk_score, topk_inds, topk_clses, topk_ys, topk_xs

def ctdet_decode(centermap, K=4):
    detections = []
    centermap=torch.sigmoid(centermap)
    centermap = _nms(centermap)

    save_center = centermap[0, 0, ...]
    save_center = save_center.cpu().detach().numpy() * 255
    save_center = save_center.astype(np.uint8)

    centermap = centermap[:, 0, :, :]
    _, inds, clses, ys, xs = _topk(centermap, K)

    xs = xs.view(1, K, 1) + 0.5
    ys = ys.view(1, K, 1) + 0.5

    xs *= 4
    ys *= 4

    y1 = ys[:, 0, :].view(1, -1)
    x1 = xs[:, 0, :].view(1, -1)
    y2 = ys[:, 1, :].view(1, -1)
    x2 = xs[:, 1, :].view(1, -1)
    y3 = ys[:, 2, :].view(1, -1)
    x3 = xs[:, 2, :].view(1, -1)
    y4 = ys[:, 3, :].view(1, -1)
    x4 = xs[:, 3, :].view(1, -1)
    detections = torch.cat((x1, y1, x2, y2, x3, y3, x4, y4), dim=1)
    detections = detections.cpu().detach().numpy()

    if len(detections) == 0:
        detections = [[0,0,0,0,0,0,0,0,0,0]]
        return np.array(detections)
    else:
        return np.concatenate(np.array(detections), 0)
def normalize(image):
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]
    mean = np.array([[mean]])
    std = np.array([[std]])
    image = (image.astype(np.float32) - mean) / std
    return image
def organize_quad_points(box):
    '''
    turn into clock-wise points: quad
    p0:lt,p1:rt,p2:rb,p4:lb
    '''
    p0,p3, p1,p2 = sorted(box, key=lambda p: p[0])
    p03=[p0,p3]
    p12=[p1,p2]
    p0, p3 = sorted(p03, key=lambda p: p[1])
    p1, p2 = sorted(p12, key=lambda p: p[1])
    return [p0,p1,p2,p3]
def crop_quad(img,box):
    p0, p1, p2, p3 = box
    (x0, y0), (x1, y1), (x2, y2), (x3, y3) = box
    w,h=((x1-x0+x2-x3)//2,(y3-y0+y2-y1)//2)
    w,h=int(w),int(h)
    M=cv2.getPerspectiveTransform(np.float32([p0,p1,p3,p2]),np.float32([[0,0],[w,0],[0,h],[w,h]]))
    img=cv2.warpPerspective(img,M,(w,h))
    return img
def img_preprocess2(image, target_shape,bboxes=None,  correct_box=False):
    """
    RGB转换 -> resize(resize不改变原图的高宽比) -> normalize
    并可以选择是否校正bbox
    :param image_org: 要处理的图像
    :param target_shape: 对图像处理后，期望得到的图像shape，存储格式为(h, w)
    :return: 处理之后的图像，shape为target_shape
    """
    h_target, w_target = target_shape
    h_org, w_org, _ = image.shape

    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB).astype(np.float32)

    resize_ratio = min(1.0 * w_target / w_org, 1.0 * h_target / h_org)
    resize_w = int(resize_ratio * w_org)
    resize_h = int(resize_ratio * h_org)
    image_resized = cv2.resize(image, (resize_w, resize_h))

    image_paded = np.full((h_target, w_target, 3), 128.0)
    dw = int((w_target - resize_w) / 2)
    dh = int((h_target - resize_h) / 2)
    image_paded[dh:resize_h+dh, dw:resize_w+dw,:] = image_resized
    image = image_paded / 255.0
    image = normalize(image)

    if correct_box:
        bboxes[:, [0, 2]] = bboxes[:, [0, 2]] * resize_ratio + dw
        bboxes[:, [1, 3]] = bboxes[:, [1, 3]] * resize_ratio + dh
        return image, bboxes
    return image,resize_ratio,dw,dh
class Rectifier:
    def __init__(self,model_path=os.path.dirname(__file__)+'/model.pth'):
        self.device=torch.device('cuda' if torch.cuda.is_available() else "cpu")
        print("device:",self.device)
        from .backbone import EfficientDetBackbone
        self.model=EfficientDetBackbone()
        self.model.to(self.device)
        state_dict=torch.load(model_path,map_location=self.device)
        self.model.load_state_dict(state_dict)
        self.model.eval()
        self.transform=transforms.Compose([
            # transforms.ToPILImage(),
            transforms.Resize((512,512)),
            transforms.ToTensor()
        ])
    def predict(self,img):
        # if not isinstance(img,Image.Image):
        #     img=Image.fromarray(img).convert('RGB')
        if isinstance(img,Image.Image):
            img=np.array(img).astype(np.float32)
            img=cv2.cvtColor(img,cv2.COLOR_RGB2BGR)
        img,ratio,dw,dh=img_preprocess2(img,target_shape=(512,512))
        # else:
        #     img=cv2.cvtColor(img,cv2.COLOR_BGR2RGB).astype(np.float32)
        # h,w=img.shape[:2]
        # rh=h/512
        # rw=w/512
        # img=cv2.resize(img,(512,512))/255
        img=torch.from_numpy(img.astype(np.float32)).transpose(1,2).transpose(0,1)
        img=img.unsqueeze(0)
        img = img.to(self.device)
        y=self.model(img)
        output = ctdet_decode(y)
        output=output.reshape((4,2))
        output[:,0]=(output[:,0]-dw)/ratio
        output[:,1]=(output[:,1]-dh)/ratio
        return output
_rectifier=None
def rectify_bxd(img):
    if isinstance(img,Image.Image):
        cvimg=np.array(img)
        cvimg=cv2.cvtColor(cvimg,cv2.COLOR_RGB2BGR)
    else:
        cvimg=img
    global _rectifier
    if not _rectifier:
        _rectifier=Rectifier()
    points=_rectifier.predict(img)
    points=organize_quad_points(points)
    img=crop_quad(cvimg,points)
    return img

def demo():
    dir="/home/ars/sda5/data/chaoyuan/datasets/detect_datasets/保险单矫正/val"
    out_dir=dir+'_out'
    import wpcv
    out=wpcv.ImageSaver(out_dir,remake_dir=True)
    out2=wpcv.ImageSaver(dir+'_out2',remake_dir=True)
    rectifier=Rectifier('/home/ars/disk/work/超远/chaoyuan-part1-dev/scripts/tests/model.pth')
    # print(rectifier.model)
    fs=glob.glob(dir+'/*.jpg')
    for i,f in enumerate(fs):
        # img=Image.open(f)
        img=cv2.imread(f)
        pilimg=Image.open(f)
        y=rectifier.predict(img)
        y=wpcv.organize_quad_points(y)
        y=np.array(y)
        # y=y.tolist()
        # print(y)
        im = wpcv.draw_polygon(pilimg,y,color='red',width=5)
        out.save(im)
        im=wpcv.crop_quad(img,y)
        out2.save(im)
        print(i,f,y)

if __name__ == '__main__':
    demo()


import torch
import cv2
import numpy as np
from PIL import Image
from detro import utils as ut
from .predictor import Predictor



class ClassicalDetectorBase(Predictor):
    MAX_OUTPUT_NUM = 100
    SCORE_THRESH = 0.1
    def heatmap2image(self,heatmap,img=None, scaleX=None,scaleY=None,pad_x=None,pad_y=None):
        scaleX=self.cache['scaleX'] if scaleX is None else scaleX
        scaleY=self.cache['scaleY'] if scaleY is None else scaleY
        pad_x=self.cache['pad_x'] if pad_x is None else pad_x
        pad_y=self.cache['pad_y'] if pad_y is None else pad_y
        img = self.cache['img']
        heatmap_downsample=self.HEATMAP_DOWNSAMPLE
        heatmap = torch.nn.functional.sigmoid(heatmap)
        heatmap *= 255
        heatmap = heatmap.detach().cpu().numpy().astype(np.uint8)
        h,w=heatmap.shape
        w*=heatmap_downsample
        h*=heatmap_downsample
        heatmap=cv2.resize(heatmap,(w,h))
        heatmap=heatmap[pad_y:h-pad_y,pad_x:w-pad_x]
        new_h,new_w=int(heatmap.shape[0]/scaleY),int(heatmap.shape[1]/scaleX)
        if img is not None:
            new_h,new_w=img.shape[:2]
        heatmap=cv2.resize(heatmap,(new_w,new_h))
        im=img.copy()
        im[...,0]+=heatmap
        heatmap=cv2.cvtColor(im,cv2.COLOR_RGB2BGR)
        return heatmap
    def preprocess(self,img):
        h, w = img.shape[:2]
        dst_w, dst_h = self.INPUT_SIZE
        scale=min(dst_w/w,dst_h/h)
        scale_x,scale_y=scale,scale
        new_h=int(scale_y*h)
        new_w=int(scale_x*w)
        pad_x=(dst_w-new_w)//2
        pad_y=(dst_h-new_h)//2
        img_padded=np.full((dst_h,dst_w,3),0)
        img_resized=cv2.resize(img,(new_w,new_h))
        img_padded[pad_y:pad_y+new_h,pad_x:pad_x+new_w]=img_resized
        img=img_padded
        img = img / 255
        img = ut.normalize(img)
        img = np.transpose(img, (2, 0, 1))
        img = np.array([img]).astype(np.float32)
        img = torch.from_numpy(img)
        img = img.to(self.DEVICE)
        return img,[scale_x,scale_y,pad_x,pad_y]
    def decode_outputs(self,outputs,down,k):
        raise NotImplementedError
    def postprocess(self,res,score_thresh):
        raise NotImplementedError
    def predict(self,img,max_output_num=None,score_thresh=None):
        if max_output_num is None:
            assert self.MAX_OUTPUT_NUM is not None
            max_output_num=self.MAX_OUTPUT_NUM
        if score_thresh is None:
            assert self.SCORE_THRESH is not None
            score_thresh=self.SCORE_THRESH
        down=self.HEATMAP_DOWNSAMPLE
        if isinstance(img,Image.Image):
            img=np.array(img).astype(np.float32)
        else:
            img=cv2.cvtColor(img,cv2.COLOR_BGR2RGB).astype(np.float32)
        self.cache.update(img=img)
        img,(scaleX,scaleY,pad_x,pad_y)=self.preprocess(img)
        outputs= self.model(img)
        self.cache.update(outputs=outputs,scaleX=scaleX,scaleY=scaleY,pad_x=pad_x,pad_y=pad_y)
        res=self.decode_outputs(outputs,down=down,k=max_output_num)[0]
        res=self.postprocess(res,score_thresh)
        return res


from detro import utils as ut
from .classical_detector import ClassicalDetectorBase
import numpy as np


def postprocess_result_circle(res, score_thresh=None, scaleX=None, scaleY=None, pad_x=None, pad_y=None):
    res[:, :2] = (res[:, :2] - np.array([pad_x, pad_y])) / np.array([scaleX, scaleY])
    res[:, 2] /= scaleX
    polygons = list(filter(lambda obj: obj[3] > score_thresh, res))
    polygons = np.array(polygons)
    return polygons


def postprocess_result_polygon(res, num_points, score_thresh=None, scaleX=None, scaleY=None, pad_x=None, pad_y=None):
    res[:, :num_points * 2] = (
            (res[:, :num_points * 2].reshape((-1, num_points, 2)) - np.array([pad_x, pad_y])) / np.array(
        [scaleX, scaleY])).reshape((-1, num_points * 2))
    polygons = list(filter(lambda polygon: polygon[num_points * 2] > score_thresh, res))
    polygons = np.array(polygons)
    return polygons


class CircleDetector(ClassicalDetectorBase):
    def decode_outputs(self, outputs, *args, **kwargs):
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_circledet(heatmap, offsets, down=self.HEATMAP_DOWNSAMPLE,
                                           k=self.MAX_OUTPUT_NUM, nms_radius=3)

    def postprocess(self, res, score_thresh=None, scaleX=None, scaleY=None, pad_x=None, pad_y=None):
        score_thresh = score_thresh if score_thresh is not None else self.SCORE_THRESH
        scaleX = self.cache['scaleX'] if scaleX is None else scaleX
        scaleY = self.cache['scaleY'] if scaleY is None else scaleY
        pad_x = self.cache['pad_x'] if pad_x is None else pad_x
        pad_y = self.cache['pad_y'] if pad_y is None else pad_y
        return postprocess_result_circle(res, score_thresh, scaleX, scaleY, pad_x, pad_y)


class PolygonDetectorBase(ClassicalDetectorBase):
    __required__ = ClassicalDetectorBase.__required__+['NUM_POINTS']
    def postprocess(self, res, score_thresh=None, scaleX=None, scaleY=None, pad_x=None, pad_y=None):
        score_thresh = score_thresh if score_thresh is not None else self.SCORE_THRESH
        scaleX = self.cache['scaleX'] if scaleX is None else scaleX
        scaleY = self.cache['scaleY'] if scaleY is None else scaleY
        pad_x = self.cache['pad_x'] if pad_x is None else pad_x
        pad_y = self.cache['pad_y'] if pad_y is None else pad_y
        return postprocess_result_polygon(res, self.NUM_POINTS, score_thresh, scaleX, scaleY, pad_x, pad_y)


class PolygonDetDetector(PolygonDetectorBase):
    def decode_outputs(self, outputs, down, k):
        num_points = self.NUM_POINTS
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_polygondet(heatmap, offsets, num_points=num_points, down=down, k=k)


class PolygonDetDetectorPolar(PolygonDetectorBase):

    def decode_outputs(self, outputs, down, k):
        num_points = self.NUM_POINTS
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_polygondet_polar(heatmap, offsets, num_points=num_points, down=down, k=k)

class FCOSDetector(PolygonDetectorBase):
    NUM_POINTS = 2
    def decode_outputs(self, outputs, down, k):
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_fcosdet(heatmap, offsets, down, k)


class CenterDetDetector(PolygonDetectorBase):
    NUM_POINTS = 2
    def decode_outputs(self, outputs, down, k):
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_centerdet(heatmap, offsets, down, k)



import cv2
from PIL import Image
import numpy as np

def cv2img(img):
    if isinstance(img,Image.Image):
        img=np.array(img)
        if len(img.shape)==3:img=img[:,:,::-1]
        return img
    return img
def pilimg(img):
    if isinstance(img,Image.Image):return img
    if isinstance(img,np.ndarray):
        if len(img.shape)==3:img=img[:,:,::-1]
    return Image.fromarray(np.array(img).astype(np.uint8))


from .dataset import FCOSDetDataset
from .network import FCOSDetCriterion,FCOSNet
from .predict import FCOSDetector
from .training import train
from .testing import test
from detro.configs import DetConfigBase
from wpcv.utils import data_aug, det_aug, img_aug, pil_ops
class FCOSDetConfig(DetConfigBase):
    DET_TYPE = 'bbox'
    FONT_PATH='/home/ars/sda6/work/ProjectBox/QuadNet/detro/data/msyh.ttf'
    INPUT_SIZE = (512, 512)
    BATCH_SIZE = 4
    HEATMAP_DOWNSAMPLE = 4
    CLASSES = ['0']
    WEIGHTS_DIR = 'weights/training'
    TEST_WEIGHTS_PATH='weights/training/model_best.pkl'
    transform = data_aug.Compose([
        det_aug.ToPILImage(),
        data_aug.RandomOrder([
            det_aug.Limitsize(600),
            # det_aug.RandomShear(20, 30),
            det_aug.RandomRotate(90),
            det_aug.RandomTranslate([600, 600]),
            det_aug.RandomVerticalFlip(),
            det_aug.RandomHorizontalFlip(),
            data_aug.Zip([
                data_aug.Compose([
                    img_aug.ColorJitter(brightness=0.1, contrast=0.1, saturation=0.1, hue=0.4),
                    img_aug.RandomApply(pil_ops.gaussian_blur, p=0.3, radius=1),
                    img_aug.RandomApply(pil_ops.sp_noise, p=0.3),
                ]),
                data_aug.Identical()
            ]),
        ]),
        det_aug.Resize(INPUT_SIZE, keep_ratio=True, fillcolor='black'),
    ])
    criterion = staticmethod(FCOSDetCriterion)
    def get_dataset(self):
        return FCOSDetDataset(self)
    def get_model(self):
        return FCOSNet(num_classes=len(self.CLASSES))
    def get_detector(self):
        return FCOSDetector(self,model=self.get_model(),WEIGHTS_PATH=self.TEST_WEIGHTS_PATH)
    def test(self):
        test(self)
    def train(self):
        train(self)
    def init(self):
        super().init()
        self.dataset=self.get_dataset()





import torch
import numpy as np
import os, glob, json, random
from torch.utils.data import Dataset
import cv2
from detro import utils as ut
from detro.configs import ConfigBase



class _HintBase:
    INPUT_SIZE = (512, 512)
    BATCH_SIZE = 4
    IMG_DIR = None
    LABEL_DIR = None
    DATA_DIRS=None
    NUM_REG_PARAMS=None
    ANNOT_FORMAT = 'voc'
    ANNOT_FILE_EXT = '.xml'
    NUM_POINTS = None
    CLASSES = None
    NUM_CLASSES = None
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    transform = None
    CLS2IDX = None
    @classmethod
    def get_options(cls):
        dic = {}
        for k, v in cls.__dict__.items():
            if not (k.startswith('__') and k.endswith('__') and k not in ['init']):
                dic[k] = v
        return dic

class DetDatasetBase(_HintBase):
    __allowed_options__=list(_HintBase.get_options().keys())
    __required__=['BATCH_SIZE','INPUT_SIZE']
    def __init__(self,cfg=None):
        self.cfg = cfg
        if cfg:
            assert isinstance(cfg, ConfigBase)
            self.set_options(cfg)
        self.init()
        self.check_required()
    def set_options(self,cfg):
        for k, v in cfg.to_dict().items():
            if k in self.__allowed_options__:
                if v is None:
                    v_old=getattr(self,k,None)
                    v=v_old
                setattr(self, k, v)
    def check_required(self):
        for key in self.__required__:
            if getattr(self, key, None) is None:
                raise Exception('Argument %s is required but not given.' % (key))
    def init(self):
        pass


class DetDataset(DetDatasetBase):
    __required__ = ['BATCH_SIZE', 'INPUT_SIZE','CLS2IDX','DATA_DIRS','DEVICE']
    def init(self):
        if self.CLASSES:
            self.CLS2IDX = {cls: idx for idx, cls in enumerate(self.CLASSES)}
            self.NUM_CLASSES = len(self.CLASSES)
        else:
            self.CLS2IDX = None
        if self.IMG_DIR:
            self.DATA_DIRS = self.DATA_DIRS or []
            # data_dirs format: (dir,dir, labelme, .json)
            self.DATA_DIRS.append((self.IMG_DIR, self.LABEL_DIR, self.ANNOT_FORMAT, self.ANNOT_FILE_EXT))
        self.annotations = self.load_annotations_multi_dirs(self.DATA_DIRS)
        random.shuffle(self.annotations)
        self.current_index = -1
        self.num_batches = int(len(self.annotations) / self.BATCH_SIZE)
    def set_device(self, device):
        self.DEVICE = device

    def process_annotation(self, annot):
        raise NotImplementedError

    def process_batch_annots(self, batch_annots):
        raise NotImplementedError

    def create_label(self, img, shapes):
        raise NotImplementedError

    def __iter__(self):
        return self

    def __next__(self):
        self.current_index += 1
        if self.current_index == self.num_batches:
            self.current_index = -1
            random.shuffle(self.annotations)
            raise StopIteration
        batch_annots = self.annotations[self.current_index * self.BATCH_SIZE:(self.current_index + 1) * self.BATCH_SIZE]
        batch_inputs, batch_labels = self.process_batch_annots(batch_annots)
        return batch_inputs, batch_labels

    def __len__(self):
        return len(self.annotations)

    def load_annot_file_labelme(self, label_path):
        objs = []
        with open(label_path, 'r') as fp:
            annot = json.load(fp)
            shapes = annot['shapes']
            if not shapes:  # 没有标注框的就不要
                return objs
            for shape in shapes:
                points = shape['points']
                label = self.CLS2IDX[shape['label']]
                objs.append([points, label])
        return objs

    def load_annot_file_pascalvoc(self, label_path):
        annot = ut.read_pascalvoc_annotation(label_path, cls2idx=self.CLS2IDX)
        objs = [[ut.ltrb_to_quad(ltrb), cls] for ltrb, cls in annot]
        return objs

    def load_annotations_from_dir(self, img_dir, label_dir, label_ext=None, format=None):
        format = format or 'labelme'
        if not label_ext:
            if format == 'labelme':
                label_ext = '.json'
            else:
                # voc as default
                label_ext = '.xml'
        fs = []
        for s in [img_dir + '/*' + ext for ext in ['.jpg', '.png', 'jpeg', '.bmp']]:
            fs += glob.glob(s)
        annotations = []
        for i, img_path in enumerate(fs):
            label_path = label_dir + '/' + os.path.basename(img_path).rsplit('.', maxsplit=1)[0] + label_ext
            if not os.path.exists(label_path):
                continue
            if format == 'labelme':
                objs = self.load_annot_file_labelme(label_path)
            else:
                objs = self.load_annot_file_pascalvoc(label_path)
            annotations.append((img_path, objs))
        return annotations

    def load_annotations_multi_dirs(self, data_dirs):
        annotations = []
        for img_dir, label_dir, annot_format, annot_file_ext in data_dirs:
            annotations += self.load_annotations_from_dir(img_dir, label_dir, format=annot_format,
                                                          label_ext=annot_file_ext)
        return annotations



import wpcv
import math
import numpy as np

import math
import numpy as np
from .polygon_utils import organize_quad_points,rotate_points,bounding_rect
##########################Box Operations#############################
def resize_box(box,size):
    cx,cy,w,h=ltrb_to_ccwh(box)
    nw,nh=size
    l=cx-nw//2
    r=cx+nw//2
    t=cy-nh//2
    b=cy+nh//2
    return [l,t,r,b]
def rescale_box(box,scale):
    '''rescale with  image'''
    if isinstance(scale,(tuple,list)):
        scaleX,scaleY=scale
    else:
        scaleX=scaleY=scale
    l,t,r,b=box
    l*=scaleX
    r*=scaleX
    t*=scaleY
    b*=scaleY
    return [l,t,r,b]


def rotate_boxes(boxes,angle,cx,cy,h,w):
    '''原创'''
    boxes=[list(box) for box in boxes]
    quads = [ltrb_to_quad(box) for box in boxes]
    quads=rotate_points(quads,angle,cx,cy,h,w)
    boxes=[bounding_rect(quad) for quad in quads]
    boxes=np.array(boxes)
    return boxes
def rotate_box(box, angle, cx, cy, h, w):
    return rotate_boxes([box],angle,cx,cy,h,w)
def shift_box(box,offset,limits=None):
    l,t,r,b=box
    ofx,ofy=offset
    l+=ofx
    r+=ofx
    t+=ofy
    b+=ofy
    return limit_box([l,t,r,b],limits=limits)
def translate_box(box,offset,limits=None):
    return shift_box(box,offset,limits)
def flip_box_horizontal(box,imsize):
    imw,imh=imsize
    l,t,r,b=box
    l=imw-r
    r=imw-l
    return [l,t,r,b]
def flip_box_vertical(box,imsize):
    imw,imh=imsize
    l,t,r,b=box
    t=imh-b
    b=imh-t
    return [l,t,r,b]
def pad_box(box,pad_size=5,pad_ratio=None,limits=None):
    from math import inf
    limits=limits or (-inf,-inf,inf,inf)
    l,t,r,b=box
    bh=b-t
    bw=r-l
    if pad_ratio is not None:
        if isinstance(pad_ratio,(tuple,list)):
            if len(pad_ratio)==2:
                pad_l=pad_r=int(min(bh,bw)*pad_ratio[0])
                pad_t=pad_b=int(min(bh,bw)*pad_ratio[1])
            else:
                assert len(pad_ratio)==4
                pad_l= int(min(bh, bw) * pad_ratio[0])
                pad_t= int(min(bh, bw) * pad_ratio[1])
                pad_r= int(min(bh, bw) * pad_ratio[2])
                pad_b= int(min(bh, bw) * pad_ratio[3])
        else:
            pad_l=pad_t=pad_r=pad_b=int(min(bh,bw)*pad_ratio)
    else:
        if isinstance(pad_size,(tuple,list)):
            if len(pad_size)==2:
                pad_l=pad_r=pad_size[0]
                pad_t=pad_b=pad_size[1]
            else:
                assert len(pad_size)==4
                pad_l,pad_t,pad_r,pad_b=pad_size
        else:
            pad_l=pad_t=pad_r=pad_b=pad_size
    l-=pad_l
    t-=pad_t
    r+=pad_r
    b+=pad_b
    return limit_box([l,t,r,b],limits=limits)

def limit_box(box,limits=None):
    if limits is None:return box
    if len(limits)==2:
        ml,mt=0,0
        mr,mb=limits
    else:
        assert len(limits)==4
        ml,mt,mr,mb=limits
    l,t,r,b=box
    l=max(ml,l)
    t=max(mt,t)
    r=min(mr,r)
    b=min(mb,b)
    if l>=r:
        return None
    if t>=b:return None
    return [l,t,r,b]
############################Sort Operations#############################
def min_enclosing_box(boxes):
    boxes=np.array(boxes)
    ml=np.min(boxes[:,0])
    mt=np.min(boxes[:,1])
    mr=np.max(boxes[:,2])
    mb=np.max(boxes[:,3])
    return [ml,mt,mr,mb]
def box_in_area(box,size):
    l,t,r,b=box
    mr,mb=size
    if l<0 or t<0 or r>mr or b>mb:
        return False
    else:
        return True
############################ConvertingTools#############################
def ltrb_to_quad(box):
    l,t,r,b=box
    quad=[[l,t],[r,t],[r,b],[l,b]]
    return quad

def ltrb_to_ccwh(box):
    l, t, r, b = box
    w = r - l
    h = b - t
    cx = (l + r) // 2
    cy = (t + b) // 2
    return [cx,cy,w,h]
def ccwh_to_ltrb(box):
    cx,cy,w,h=box
    l=cx-w//2
    t=cy-h//2
    r=cx+w//2
    b=cy+h//2
    return [l,t,r,b]
def oowh_to_ltrb(box):
    '''o:origin'''
    x, y, w, h = box
    return [int(i) for i in (x, y, x + w, y + h)]
def ltrb_to_oowh(box):
    l, t, r, b = box
    return [l, t, r - l, b - t]
def xywh_to_ltrb(box):
    '''same with oowh'''
    x,y,w,h=box
    return [int(i) for i in (x,y,x+w,y+h)]
def ltrb_to_xywh(box):
    l,t,r,b=box
    return [l,t,r-l,b-t]

def quad_to_four_corners(quad):
    (x0, y0), (x1, y1), (x2, y2), (x3, y3) = quad
    return [x0,y0,x1,y1,x2,y2,x3,y3]
def four_corners_to_quad(box):
    x0,y0,x1,y1,x2,y2,x3,y3 = box
    return [(x0, y0), (x1, y1), (x2, y2), (x3, y3)]
###########################Calculations##############################
def calc_iou(box1,box2):
    l1,t1,r1,b1=box1
    l2,t2,r2,b2=box2
    w1,h1=r1-l1,b1-t1
    w2,h2=r2-l2,b2-t2
    width=min(r1,r2)-max(l1,l2)
    height=min(b1,b2)-max(t1,t2)
    width=max(width,0)
    height=max(height,0)
    area=width*height
    return area/(w1*h1+w2*h2-area)
def calc_iou2(box1,box2):
    width=min(box1[2],box2[2])-max(box1[0],box2[0])
    height=min(box1[3],box2[3])-max(box1[1],box2[1])
    width=max(width,0)
    height=max(height,0)
    area=width*height
    return area/((box1[2]-box1[0])*(box1[3]-box1[1])+(box2[2]-box2[0])*(box2[3]-box2[1])-area)
def calc_iou_batch_with_batch(boxes1,boxes2):
    box1=boxes1
    box2=boxes2
    import numpy as np
    box1=np.array(box1)
    box2=np.array(box2)
    width = np.min(np.vstack([box1[...,2], box2[...,2]]),axis=0) - np.max(np.vstack([box1[...,0], box2[...,0]]),axis=0)
    height = np.min(np.vstack([box1[...,3], box2[...,3]]),axis=0) - np.max(np.vstack([box1[...,1], box2[...,1]]),axis=0)
    width = np.clip(width,0,np.inf)
    height = np.clip(height,0,np.inf)
    area = width * height
    return area / ((box1[...,2] - box1[...,0]) * (box1[...,3] - box1[...,1]) + (box2[...,2] - box2[...,0]) * (box2[...,3] - box2[...,1]) - area)
def calc_iou_batch(boxes,box):
    '''calculate iou between batch bboxes with one bbox , 原创'''
    box1=boxes
    box2=box
    import numpy as np
    box1 = np.array(box1)
    box2 = np.array(box2)
    shape=box1.shape
    width = np.min(np.vstack([box1[..., 2], np.full(shape[:-1],box2[2])]), axis=0) - np.max(np.vstack([box1[..., 0], np.full(shape[:-1],box2[0])]),axis=0)
    height = np.min(np.vstack([box1[..., 3], np.full(shape[:-1],box2[3])]), axis=0) - np.max(np.vstack([box1[..., 1], np.full(shape[:-1],box2[1])]),axis=0)
    width = np.clip(width, 0, np.inf)
    height = np.clip(height, 0, np.inf)
    area = width * height
    return area / ((box1[...,2] - box1[...,0]) * (box1[...,3] - box1[...,1]) + (box2[2] - box2[0]) * (box2[3] - box2[1]) - area)

def calc_box_area(box):
    l,t,r,b=box
    return (r-l)*(b-t)





from .label_utils import *
from .myutils import *
from .font_utils import *
from .annot_utils import *
from .box_utils import *
from .polygon_utils import *
from .decode_utils import *
from .visualize import *
from .img_utils import *


from .predictor import *
from .classical_detector import ClassicalDetectorBase
from .shapedet_detector import PolygonDetectorBase,CircleDetector,PolygonDetDetector,PolygonDetDetectorPolar,CenterDetDetector,FCOSDetector



from wpcv import *
import wpcv


def heatmap2image(heatmap):
    heatmap = np.array(heatmap) * 255
    heatmap = heatmap.astype(np.uint8)
    return heatmap
_saver=None
def get_default_saver():
    global _saver
    if _saver is None:
        _saver = ImageSaver('debug/default', remake_dir=True, auto_make_subdir=True)
    return _saver
def save(im):
    saver=get_default_saver()
    return saver.save(im)


import numpy as np
import cv2
from PIL import Image
from detro import utils as ut
def visualize(img,res):
    img=ut.pilimg(img)
    for circle in res:
        text='%.2f'%(circle[3])
        ut.draw_circle(img,center=circle[:2],radius=circle[2],text=text,fillcolor='blue')
    return img


from PIL import ImageFont
import os

_root = os.path.dirname(__file__)
_default_font_path = _root + '/../data/msyh.ttf'


def set_default_font_path(path):
    global _default_font_path
    _default_font_path = path


def get_default_font(fontsize=16):
    font = _default_font_path
    return ImageFont.truetype(font, size=fontsize)



import torch
import torch.nn as nn
import cv2
import numpy as np

from efficientdet.utils import BBoxTransform, ClipBoxes
from utils.utils import postprocess, invert_affine, display


def calc_iou(a, b):
    # a(anchor) [boxes, (y1, x1, y2, x2)]
    # b(gt, coco-style) [boxes, (x1, y1, x2, y2)]

    area = (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])
    iw = torch.min(torch.unsqueeze(a[:, 3], dim=1), b[:, 2]) - torch.max(torch.unsqueeze(a[:, 1], 1), b[:, 0])
    ih = torch.min(torch.unsqueeze(a[:, 2], dim=1), b[:, 3]) - torch.max(torch.unsqueeze(a[:, 0], 1), b[:, 1])
    iw = torch.clamp(iw, min=0)
    ih = torch.clamp(ih, min=0)
    ua = torch.unsqueeze((a[:, 2] - a[:, 0]) * (a[:, 3] - a[:, 1]), dim=1) + area - iw * ih
    ua = torch.clamp(ua, min=1e-8)
    intersection = iw * ih
    IoU = intersection / ua

    return IoU


class FocalLoss(nn.Module):
    def __init__(self):
        super(FocalLoss, self).__init__()

    def forward(self, classifications, regressions, anchors, annotations, **kwargs):
        alpha = 0.25
        gamma = 2.0
        batch_size = classifications.shape[0]
        classification_losses = []
        regression_losses = []

        anchor = anchors[0, :, :]  # assuming all image sizes are the same, which it is
        dtype = anchors.dtype

        anchor_widths = anchor[:, 3] - anchor[:, 1]
        anchor_heights = anchor[:, 2] - anchor[:, 0]
        anchor_ctr_x = anchor[:, 1] + 0.5 * anchor_widths
        anchor_ctr_y = anchor[:, 0] + 0.5 * anchor_heights

        for j in range(batch_size):

            classification = classifications[j, :, :]
            regression = regressions[j, :, :]

            bbox_annotation = annotations[j]
            bbox_annotation = bbox_annotation[bbox_annotation[:, 4] != -1]

            classification = torch.clamp(classification, 1e-4, 1.0 - 1e-4)
            
            if bbox_annotation.shape[0] == 0:
                if torch.cuda.is_available():
                    
                    alpha_factor = torch.ones_like(classification) * alpha
                    alpha_factor = alpha_factor.cuda()
                    alpha_factor = 1. - alpha_factor
                    focal_weight = classification
                    focal_weight = alpha_factor * torch.pow(focal_weight, gamma)
                    
                    bce = -(torch.log(1.0 - classification))
                    
                    cls_loss = focal_weight * bce
                    
                    regression_losses.append(torch.tensor(0).to(dtype).cuda())
                    classification_losses.append(cls_loss.sum())
                else:
                    
                    alpha_factor = torch.ones_like(classification) * alpha
                    alpha_factor = 1. - alpha_factor
                    focal_weight = classification
                    focal_weight = alpha_factor * torch.pow(focal_weight, gamma)
                    
                    bce = -(torch.log(1.0 - classification))
                    
                    cls_loss = focal_weight * bce
                    
                    regression_losses.append(torch.tensor(0).to(dtype))
                    classification_losses.append(cls_loss.sum())

                continue
                
            IoU = calc_iou(anchor[:, :], bbox_annotation[:, :4])

            IoU_max, IoU_argmax = torch.max(IoU, dim=1)

            # compute the loss for classification
            targets = torch.ones_like(classification) * -1
            if torch.cuda.is_available():
                targets = targets.cuda()

            targets[torch.lt(IoU_max, 0.4), :] = 0

            positive_indices = torch.ge(IoU_max, 0.5)

            num_positive_anchors = positive_indices.sum()

            assigned_annotations = bbox_annotation[IoU_argmax, :]

            targets[positive_indices, :] = 0
            targets[positive_indices, assigned_annotations[positive_indices, 4].long()] = 1

            alpha_factor = torch.ones_like(targets) * alpha
            if torch.cuda.is_available():
                alpha_factor = alpha_factor.cuda()

            alpha_factor = torch.where(torch.eq(targets, 1.), alpha_factor, 1. - alpha_factor)
            focal_weight = torch.where(torch.eq(targets, 1.), 1. - classification, classification)
            focal_weight = alpha_factor * torch.pow(focal_weight, gamma)

            bce = -(targets * torch.log(classification) + (1.0 - targets) * torch.log(1.0 - classification))

            cls_loss = focal_weight * bce

            zeros = torch.zeros_like(cls_loss)
            if torch.cuda.is_available():
                zeros = zeros.cuda()
            cls_loss = torch.where(torch.ne(targets, -1.0), cls_loss, zeros)

            classification_losses.append(cls_loss.sum() / torch.clamp(num_positive_anchors.to(dtype), min=1.0))

            if positive_indices.sum() > 0:
                assigned_annotations = assigned_annotations[positive_indices, :]

                anchor_widths_pi = anchor_widths[positive_indices]
                anchor_heights_pi = anchor_heights[positive_indices]
                anchor_ctr_x_pi = anchor_ctr_x[positive_indices]
                anchor_ctr_y_pi = anchor_ctr_y[positive_indices]

                gt_widths = assigned_annotations[:, 2] - assigned_annotations[:, 0]
                gt_heights = assigned_annotations[:, 3] - assigned_annotations[:, 1]
                gt_ctr_x = assigned_annotations[:, 0] + 0.5 * gt_widths
                gt_ctr_y = assigned_annotations[:, 1] + 0.5 * gt_heights

                # efficientdet style
                gt_widths = torch.clamp(gt_widths, min=1)
                gt_heights = torch.clamp(gt_heights, min=1)

                targets_dx = (gt_ctr_x - anchor_ctr_x_pi) / anchor_widths_pi
                targets_dy = (gt_ctr_y - anchor_ctr_y_pi) / anchor_heights_pi
                targets_dw = torch.log(gt_widths / anchor_widths_pi)
                targets_dh = torch.log(gt_heights / anchor_heights_pi)

                targets = torch.stack((targets_dy, targets_dx, targets_dh, targets_dw))
                targets = targets.t()

                regression_diff = torch.abs(targets - regression[positive_indices, :])

                regression_loss = torch.where(
                    torch.le(regression_diff, 1.0 / 9.0),
                    0.5 * 9.0 * torch.pow(regression_diff, 2),
                    regression_diff - 0.5 / 9.0
                )
                regression_losses.append(regression_loss.mean())
            else:
                if torch.cuda.is_available():
                    regression_losses.append(torch.tensor(0).to(dtype).cuda())
                else:
                    regression_losses.append(torch.tensor(0).to(dtype))

        # debug
        imgs = kwargs.get('imgs', None)
        if imgs is not None:
            regressBoxes = BBoxTransform()
            clipBoxes = ClipBoxes()
            obj_list = kwargs.get('obj_list', None)
            out = postprocess(imgs.detach(),
                              torch.stack([anchors[0]] * imgs.shape[0], 0).detach(), regressions.detach(), classifications.detach(),
                              regressBoxes, clipBoxes,
                              0.5, 0.3)
            imgs = imgs.permute(0, 2, 3, 1).cpu().numpy()
            imgs = ((imgs * [0.229, 0.224, 0.225] + [0.485, 0.456, 0.406]) * 255).astype(np.uint8)
            imgs = [cv2.cvtColor(img, cv2.COLOR_RGB2BGR) for img in imgs]
            display(out, imgs, obj_list, imshow=False, imwrite=True)

        return torch.stack(classification_losses).mean(dim=0, keepdim=True), \
               torch.stack(regression_losses).mean(dim=0, keepdim=True)



import torch.nn as nn
import torch
from .resnet import BasicBlock,Bottleneck


class SeparableConvBlock(nn.Module):
    """
    created by Zylo117
    """

    def __init__(self, in_channels, out_channels=None, norm=True, activation=False, onnx_export=False):
        super(SeparableConvBlock, self).__init__()
        if out_channels is None:
            out_channels = in_channels

        # Q: whether separate conv
        #  share bias between depthwise_conv and pointwise_conv
        #  or just pointwise_conv apply bias.
        # A: Confirmed, just pointwise_conv applies bias, depthwise_conv has no bias.

        self.depthwise_conv = nn.Conv2d(in_channels, in_channels,
                                                      kernel_size=3, stride=1, groups=in_channels, bias=False)
        self.pointwise_conv = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1)

        self.norm = norm
        if self.norm:
            # Warning: pytorch momentum is different from tensorflow's, momentum_pytorch = 1 - momentum_tensorflow
            self.bn = nn.BatchNorm2d(num_features=out_channels, momentum=0.01, eps=1e-3)

        self.activation = activation
        if self.activation:
            self.relu=nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.depthwise_conv(x)
        x = self.pointwise_conv(x)

        if self.norm:
            x = self.bn(x)

        if self.activation:
            x = self.relu(x)

        return x


class BiFPN(nn.Module):
    """
    modified by Zylo117
    """

    def __init__(self, num_channels, conv_channels, first_time=False, epsilon=1e-4, onnx_export=True, attention=True):
        """
        Args:
            num_channels:
            conv_channels:
            first_time: whether the input comes directly from the efficientnet,
                        if True, downchannel it first, and downsample P5 to generate P6 then P7
            epsilon: epsilon of fast weighted attention sum of BiFPN, not the BN's epsilon
            onnx_export: if True, use Swish instead of MemoryEfficientSwish
        """
        super(BiFPN, self).__init__()
        self.epsilon = epsilon
        # Conv layers
        self.conv6_up = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv5_up = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv4_up = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv3_up = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv4_down = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv5_down = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv6_down = SeparableConvBlock(num_channels, onnx_export=onnx_export)
        self.conv7_down = SeparableConvBlock(num_channels, onnx_export=onnx_export)

        # Feature scaling layers
        self.p6_upsample = nn.Upsample(scale_factor=2, mode='nearest')
        self.p5_upsample = nn.Upsample(scale_factor=2, mode='nearest')
        self.p4_upsample = nn.Upsample(scale_factor=2, mode='nearest')
        self.p3_upsample = nn.Upsample(scale_factor=2, mode='nearest')

        self.p4_downsample = nn.MaxPool2d(3, 2)
        self.p5_downsample = nn.MaxPool2d(3, 2)
        self.p6_downsample = nn.MaxPool2d(3, 2)
        self.p7_downsample = nn.MaxPool2d(3, 2)

        self.swish = nn.ReLU(inplace=True)

        self.first_time = first_time
        if self.first_time:
            self.p5_down_channel = nn.Sequential(
                nn.Conv2d(conv_channels[2], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
            )
            self.p4_down_channel = nn.Sequential(
                nn.Conv2d(conv_channels[1], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
            )
            self.p3_down_channel = nn.Sequential(
                nn.Conv2d(conv_channels[0], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
            )

            self.p5_to_p6 = nn.Sequential(
                nn.Conv2d(conv_channels[2], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
                nn.MaxPool2d(3, 2)
            )
            self.p6_to_p7 = nn.Sequential(
                nn.MaxPool2d(3, 2)
            )

            self.p4_down_channel_2 = nn.Sequential(
                nn.Conv2d(conv_channels[1], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
            )
            self.p5_down_channel_2 = nn.Sequential(
                nn.Conv2d(conv_channels[2], num_channels, 1),
                nn.BatchNorm2d(num_channels, momentum=0.01, eps=1e-3),
            )


        # Weight
        self.p6_w1 = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.p6_w1_relu = nn.ReLU()
        self.p5_w1 = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.p5_w1_relu = nn.ReLU()
        self.p4_w1 = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.p4_w1_relu = nn.ReLU()
        self.p3_w1 = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.p3_w1_relu = nn.ReLU()

        self.p4_w2 = nn.Parameter(torch.ones(3, dtype=torch.float32), requires_grad=True)
        self.p4_w2_relu = nn.ReLU()
        self.p5_w2 = nn.Parameter(torch.ones(3, dtype=torch.float32), requires_grad=True)
        self.p5_w2_relu = nn.ReLU()
        self.p6_w2 = nn.Parameter(torch.ones(3, dtype=torch.float32), requires_grad=True)
        self.p6_w2_relu = nn.ReLU()
        self.p7_w2 = nn.Parameter(torch.ones(2, dtype=torch.float32), requires_grad=True)
        self.p7_w2_relu = nn.ReLU()

        self.attention = attention

    def forward(self, inputs):
        """
        illustration of a minimal bifpn unit
            P7_0 -------------------------> P7_2 -------->
               |-------------|                ↑
                             ↓                |
            P6_0 ---------> P6_1 ---------> P6_2 -------->
               |-------------|--------------↑ ↑
                             ↓                |
            P5_0 ---------> P5_1 ---------> P5_2 -------->
               |-------------|--------------↑ ↑
                             ↓                |
            P4_0 ---------> P4_1 ---------> P4_2 -------->
               |-------------|--------------↑ ↑
                             |--------------↓ |
            P3_0 -------------------------> P3_2 -------->
        """

        # downsample channels using same-padding conv2d to target phase's if not the same
        # judge: same phase as target,
        # if same, pass;
        # elif earlier phase, downsample to target phase's by pooling
        # elif later phase, upsample to target phase's by nearest interpolation

        if self.attention:
            p3_out, p4_out, p5_out, p6_out, p7_out = self._forward_fast_attention(inputs)
        else:
            p3_out, p4_out, p5_out, p6_out, p7_out = self._forward(inputs)

        # p3_out = self.conv3_up(self.p3_upsample(p3_out))

        return p3_out, p4_out, p5_out, p6_out, p7_out

    def _forward_fast_attention(self, inputs):
        if self.first_time:
            p3, p4, p5 = inputs

            p6_in = self.p5_to_p6(p5)
            p7_in = self.p6_to_p7(p6_in)

            p3_in = self.p3_down_channel(p3)
            p4_in = self.p4_down_channel(p4)
            p5_in = self.p5_down_channel(p5)

        else:
            # P3_0, P4_0, P5_0, P6_0 and P7_0
            p3_in, p4_in, p5_in, p6_in, p7_in = inputs

        # P7_0 to P7_2

        # Weights for P6_0 and P7_0 to P6_1
        p6_w1 = self.p6_w1_relu(self.p6_w1)
        weight = p6_w1 / (torch.sum(p6_w1, dim=0) + self.epsilon)
        # Connections for P6_0 and P7_0 to P6_1 respectively
        p6_up = self.conv6_up(self.swish(weight[0] * p6_in + weight[1] * self.p6_upsample(p7_in)))

        # Weights for P5_0 and P6_1 to P5_1
        p5_w1 = self.p5_w1_relu(self.p5_w1)
        weight = p5_w1 / (torch.sum(p5_w1, dim=0) + self.epsilon)
        # Connections for P5_0 and P6_0 to P5_1 respectively
        p5_up = self.conv5_up(self.swish(weight[0] * p5_in + weight[1] * self.p5_upsample(p6_up)))

        # Weights for P4_0 and P5_1 to P4_1
        p4_w1 = self.p4_w1_relu(self.p4_w1)
        weight = p4_w1 / (torch.sum(p4_w1, dim=0) + self.epsilon)
        # Connections for P4_0 and P5_0 to P4_1 respectively
        p4_up = self.conv4_up(self.swish(weight[0] * p4_in + weight[1] * self.p4_upsample(p5_up)))

        # Weights for P3_0 and P4_1 to P3_2
        p3_w1 = self.p3_w1_relu(self.p3_w1)
        weight = p3_w1 / (torch.sum(p3_w1, dim=0) + self.epsilon)
        # Connections for P3_0 and P4_1 to P3_2 respectively
        p3_out = self.conv3_up(self.swish(weight[0] * p3_in + weight[1] * self.p3_upsample(p4_up)))

        if self.first_time:
            p4_in = self.p4_down_channel_2(p4)
            p5_in = self.p5_down_channel_2(p5)

        # Weights for P4_0, P4_1 and P3_2 to P4_2
        p4_w2 = self.p4_w2_relu(self.p4_w2)
        weight = p4_w2 / (torch.sum(p4_w2, dim=0) + self.epsilon)
        # Connections for P4_0, P4_1 and P3_2 to P4_2 respectively
        p4_out = self.conv4_down(
            self.swish(weight[0] * p4_in + weight[1] * p4_up + weight[2] * self.p4_downsample(p3_out)))

        # Weights for P5_0, P5_1 and P4_2 to P5_2
        p5_w2 = self.p5_w2_relu(self.p5_w2)
        weight = p5_w2 / (torch.sum(p5_w2, dim=0) + self.epsilon)
        # Connections for P5_0, P5_1 and P4_2 to P5_2 respectively
        p5_out = self.conv5_down(
            self.swish(weight[0] * p5_in + weight[1] * p5_up + weight[2] * self.p5_downsample(p4_out)))

        # Weights for P6_0, P6_1 and P5_2 to P6_2
        p6_w2 = self.p6_w2_relu(self.p6_w2)
        weight = p6_w2 / (torch.sum(p6_w2, dim=0) + self.epsilon)
        # Connections for P6_0, P6_1 and P5_2 to P6_2 respectively
        p6_out = self.conv6_down(
            self.swish(weight[0] * p6_in + weight[1] * p6_up + weight[2] * self.p6_downsample(p5_out)))

        # Weights for P7_0 and P6_2 to P7_2
        p7_w2 = self.p7_w2_relu(self.p7_w2)
        weight = p7_w2 / (torch.sum(p7_w2, dim=0) + self.epsilon)
        # Connections for P7_0 and P6_2 to P7_2
        p7_out = self.conv7_down(self.swish(weight[0] * p7_in + weight[1] * self.p7_downsample(p6_out)))

        return p3_out, p4_out, p5_out, p6_out, p7_out

    def _forward(self, inputs):
        if self.first_time:
            p3, p4, p5 = inputs

            p6_in = self.p5_to_p6(p5)
            p7_in = self.p6_to_p7(p6_in)

            p3_in = self.p3_down_channel(p3)
            p4_in = self.p4_down_channel(p4)
            p5_in = self.p5_down_channel(p5)

        else:
            # P3_0, P4_0, P5_0, P6_0 and P7_0
            p3_in, p4_in, p5_in, p6_in, p7_in = inputs

        # P7_0 to P7_2

        # Connections for P6_0 and P7_0 to P6_1 respectively
        p6_up = self.conv6_up(self.swish(p6_in + self.p6_upsample(p7_in)))

        # Connections for P5_0 and P6_1 to P5_1 respectively
        p5_up = self.conv5_up(self.swish(p5_in + self.p5_upsample(p6_up)))

        # Connections for P4_0 and P5_1 to P4_1 respectively
        p4_up = self.conv4_up(self.swish(p4_in + self.p4_upsample(p5_up)))

        # Connections for P3_0 and P4_1 to P3_2 respectively
        p3_out = self.conv3_up(self.swish(p3_in + self.p3_upsample(p4_up)))

        if self.first_time:
            p4_in = self.p4_down_channel_2(p4)
            p5_in = self.p5_down_channel_2(p5)

        # Connections for P4_0, P4_1 and P3_2 to P4_2 respectively
        p4_out = self.conv4_down(
            self.swish(p4_in + p4_up + self.p4_downsample(p3_out)))

        # Connections for P5_0, P5_1 and P4_2 to P5_2 respectively
        p5_out = self.conv5_down(
            self.swish(p5_in + p5_up + self.p5_downsample(p4_out)))

        # Connections for P6_0, P6_1 and P5_2 to P6_2 respectively
        p6_out = self.conv6_down(
            self.swish(p6_in + p6_up + self.p6_downsample(p5_out)))

        # Connections for P7_0 and P6_2 to P7_2
        p7_out = self.conv7_down(self.swish(p7_in + self.p7_downsample(p6_out)))

        return p3_out, p4_out, p5_out, p6_out, p7_out


class Coef_layer(nn.Module):
    """
    modified by Zylo117
    """

    def __init__(self, in_channels):
        super(Coef_layer, self).__init__()

        self.conv0 = nn.Conv2d(in_channels, 256, kernel_size=3, padding=1, bias=False)
        self.bn0 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv1 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv2 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv3 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.header = nn.Conv2d(256, 32, kernel_size=1, stride=1, padding=0, bias=True)
        self.header.bias.data.fill_(-2.19)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv0(x)
        x = self.bn0(x)
        x = self.relu(x)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu(x)
        x = self.header(x)

        return x


class Proto_layer(nn.Module):
    """
    modified by Zylo117
    """

    def __init__(self, in_channels):
        super(Proto_layer, self).__init__()

        self.conv0 = nn.Conv2d(in_channels, 256, kernel_size=3, padding=1, bias=False)
        self.bn0 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv1 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv2 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.conv3 = nn.Conv2d(256, 256, kernel_size=3, padding=1, bias=False)
        self.bn3 = nn.BatchNorm2d(256, momentum=0.01, eps=1e-3)
        self.header = nn.Conv2d(256, 32, kernel_size=1, stride=1, padding=0, bias=True)
        self.header.bias.data.fill_(-2.19)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv0(x)
        x = self.bn0(x)
        x = self.relu(x)
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        x = self.relu(x)
        x = self.conv3(x)
        x = self.bn3(x)
        x = self.relu(x)
        x = self.header(x)

        return x



class Heatmap_layer(nn.Module):
    def __init__(self, in_channels,out_channels):
        super(Heatmap_layer, self).__init__()

        self.conv = nn.Conv2d(in_channels, 256, kernel_size=3, padding=1, bias=True)
        self.header = nn.Conv2d(256, out_channels, kernel_size=1, stride=1, padding=0, bias=True)
        self.header.bias.data.fill_(-2.19)
        self.relu=nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.relu(x)
        x = self.header(x)

        return x


class Center_layer(nn.Module):
    def __init__(self, in_channels,num_classes):
        super(Center_layer, self).__init__()

        self.conv = nn.Conv2d(in_channels, 256, kernel_size=3, padding=1, bias=True)
        self.header = nn.Conv2d(256, num_classes, kernel_size=1, stride=1, padding=0, bias=True)
        self.header.bias.data.fill_(-2.19)
        self.relu=nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.relu(x)
        x = self.header(x)

        return x


class Offset_layer(nn.Module):
    """
    modified by Zylo117
    """

    def __init__(self, in_channels):
        super().__init__()

        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1, bias=True)
        self.header = nn.Conv2d(in_channels, 8, kernel_size=1, stride=1, padding=0, bias=True)
        self.relu=nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.relu(x)
        x = self.header(x)

        return x

class Reg_layer(nn.Module):
    """
    modified by Zylo117
    """

    def __init__(self, in_channels,out_channels):
        super().__init__()

        self.conv = nn.Conv2d(in_channels, in_channels, kernel_size=3, padding=1, bias=True)
        self.header = nn.Conv2d(in_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=True)
        self.relu=nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.conv(x)
        x = self.relu(x)
        x = self.header(x)
        return x




import torch
import numpy as np
import os, glob, json, random
from torch.utils.data import Dataset
import cv2
from detro import utils as ut
from detro.dataset.det_dataset import DetDataset




class PointsDetDataset(DetDataset):

	def __init__(self, img_dir=None, label_dir=None, data_dirs=None, num_classes=None, num_points=None, classes=None,
	             batch_size=8,
	             input_size=(512, 512), params=None, transform=None, device=None):
		default_params = dict(
			heatmap_downsample=4,
			points_transform='none'
		)
		default_params.update(**(params or {}))
		params = default_params

		super().__init__(
			img_dir=img_dir, label_dir=label_dir, data_dirs=data_dirs, num_classes=num_classes, classes=classes,
			batch_size=batch_size, input_size=input_size, transform=transform, params=params, device=device,
		)
		self.num_points = num_points or self.params.get('num_points') or 4

	def data_augmentation(self, img, objs):
		objs.sort(key=lambda obj: np.array(obj[0])[:, 0].mean())
		shapes = [ut.organize_polygon_points(shape) for shape, cls in objs]
		clses = [cls for shape, cls in objs]
		if self.transform:
			img, shapes = self.transform(img, shapes)
		objs = list(zip(shapes, clses))
		return img, objs

	def transform_points(self, points):
		points_transform = self.params['points_transform'].split(',')
		points = ut.organize_polygon_points(points)
		points = np.array(points)
		center = points.mean(axis=0)
		offsets = points - center
		if 'none' in points_transform:
			return points
		if 'cartesian' in points_transform:
			offsets = offsets.reshape((-1))
		elif 'polar' in points_transform:
			offsets = ut.cartesian2polar(offsets, to_degree=True).reshape((-1))
		if 'abs' in points_transform:
			offsets = np.abs(offsets)
		return offsets

	def process_batch_annots(self, batch_annots):
		down = self.params['heatmap_downsample']
		batch_size = len(batch_annots)
		imw, imh = self.input_size
		hm_h, hm_w = int(imh / down), int(imw / down)
		batch_imgs = np.zeros((batch_size, 3, imh, imw), dtype=np.float32)
		batch_center_heatmap = np.zeros((batch_size, self.num_classes, hm_h, hm_w), dtype=np.float32)
		batch_corner_heatmap = np.zeros((batch_size, self.num_classes, hm_h, hm_w), dtype=np.float32)
		batch_offsets = np.zeros((batch_size, self.num_points * 2, hm_h, hm_w), dtype=np.float32)
		batch_offsets_mask = np.zeros((batch_size, 1, hm_h, hm_w), dtype=np.float32)
		for i, (img_path, objs) in enumerate(batch_annots):
			img = cv2.imread(img_path)
			img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
			img, objs = self.data_augmentation(img, objs)
			img = np.array(img)
			img = img / 255
			img = ut.normalize(img)
			batch_imgs[i, ...] = np.transpose(img, (2, 0, 1))
			for points, cls in objs:
				points=ut.organize_polygon_points(points)
				points = np.array(points)
				offsets=(points-points.mean(axis=0)).reshape((-1))
				cx,cy=center=(points.mean(axis=0)/down).astype(np.int)
				points=(points/down).astype(np.int)
				ut.draw_points_heatmap(batch_center_heatmap[i, cls], [center])
				batch_offsets[i,:,cy,cx]=offsets/down
				batch_offsets_mask[i,0,cy,cx]=1
				for pnt in points:
					ut.draw_points_heatmap(batch_corner_heatmap[i, cls], [pnt])
		batch_imgs = torch.from_numpy(batch_imgs).to(self.device)
		batch_center_heatmap = torch.from_numpy(batch_center_heatmap).to(self.device)
		batch_corner_heatmap = torch.from_numpy(batch_corner_heatmap).to(self.device)
		batch_offsets = torch.from_numpy(batch_offsets).to(self.device)
		batch_offsets_mask = torch.from_numpy(batch_offsets_mask).to(self.device)
		batch_labels = dict(
			center_heatmap=batch_center_heatmap,
			corner_heatmap=batch_corner_heatmap,
			offsets=batch_offsets,
			offsets_mask=batch_offsets_mask,
		)
		return batch_imgs, batch_labels



import cv2
from PIL import Image,ImageDraw

def draw_circle(img,center,radius,fillcolor='red',text=None,font=None,text_color=None):
    x,y=center=(int(center[0]),int(center[1]))
    r=radius
    draw = ImageDraw.Draw(img)
    draw.ellipse((x - r, y - r, x + r, y + r), fill=None,outline=fillcolor)
    if text:
        text_color=text_color or fillcolor
        draw.text(center,text,font=font,fill=text_color)
    return img


import torch
import numpy as np
import os
import time
import logging

def load_weights(model, weights_path=None, except_keys=[]):
    if weights_path and os.path.exists(weights_path):
        try:
            state_dict = torch.load(weights_path)
            remove_keys = []
            for k in state_dict.keys():
                for key in except_keys:
                    if k.startswith(key):
                        remove_keys.append(k)
            for k in remove_keys:
                del state_dict[k]
                print('ignore key:', k)
            model.load_state_dict(state_dict, strict=False)
        except:
            logging.warning('Cannot load pretrained model %s' % (weights_path))
            raise
    return model


def train(cfg=None):
    t = time.time()
    time_stamp = time.strftime('%m%d-%H', time.localtime(t))
    model = cfg.get_model()
    # model = load_weights(model, cfg.WEIGHTS_INIT, except_keys=['reg_layer'])
    model = load_weights(model, cfg.WEIGHTS_INIT)
    device = cfg.DEVICE
    model.to(device)
    train_data = cfg.dataset
    optimizer = torch.optim.Adam(model.parameters(), cfg.LEARN_RATE_INIT)
    lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, cfg.MAX_EPOCHS, cfg.LEARN_RATE_END)
    min_loss = 1e+10
    for epoch in range(cfg.MAX_EPOCHS):
        model.train()
        loss_history = {}
        loss_avg_dict = {}
        loss_summary = []

        for batch_image, batch_labels in train_data:
            preds = model(batch_image)
            loss_dict = cfg.criterion(preds, batch_labels)
            loss = loss_dict['loss']
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            for k, v in loss_dict.items():
                if not k in loss_history.keys():
                    loss_history[k] = []
                loss_history[k].append(v.item())
        lr_scheduler.step()
        for k, v in loss_history.items():
            loss_avg_dict[k] = np.array(v).mean()
            loss_summary.append('%s:%.4f' % (k, loss_avg_dict[k]))
        loss_epoch = loss_avg_dict['loss']
        loss_summary = '\t'.join(loss_summary)
        print('Epoch: %s\tLearningRate:%.6f\t%s' % (
            epoch, optimizer.state_dict()['param_groups'][0]['lr'], loss_summary))
        if epoch and cfg.WEIGHTS_SAVE_INTERVAL and epoch % cfg.WEIGHTS_SAVE_INTERVAL == 0:
            torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_{time_stamp}_{epoch}_loss_{loss_epoch:.4f}.pkl')
        if loss_epoch < min_loss:
            min_loss = loss_epoch
            torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_best_{time_stamp}.pkl')
            torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model_best.pkl')
        torch.save(model.state_dict(), cfg.WEIGHTS_DIR + f'/model.pkl')


if __name__ == '__main__':
    train()



import torch
from detro.configs import ConfigBase
class PredictConfig(ConfigBase):
    INPUT_SIZE=None
    CLASSES=None
    NUM_CLASSES=None
    NUM_POINTS=None
    HEATMAP_DOWNSAMPLE=None
    MAX_OUTPUT_NUM=None
    SCORE_THRESH=None
    DEVICE=None
    WEIGHTS_PATH=None
    model=None
    transform=None

    def get_model(self):
        raise
    def init(self):
        if not self.model:
            self.model=self.get_model()
        self.model.load_state_dict(torch.load(self.WEIGHTS_PATH))
        self.model.to(self.DEVICE)
        self.model.eval()
class _HintBase:
    INPUT_SIZE = None
    CLASSES = None
    NUM_CLASSES = None
    NUM_POINTS = None
    HEATMAP_DOWNSAMPLE = None
    MAX_OUTPUT_NUM = None
    SCORE_THRESH = None
    DEVICE = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    WEIGHTS_PATH = None
    model = None
    transform = None
    @classmethod
    def get_options(cls):
        dic = {}
        for k, v in cls.__dict__.items():
            if not (k.startswith('__') and k.endswith('__') and k not in ['init']):
                dic[k] = v
        return dic

class PredictorBase(_HintBase):
    __allowed_options__=_HintBase.get_options()
    __required__=['model','INPUT_SIZE','DEVICE']
    def __init__(self, cfg=None,**kwargs):
        self.cache={}
        if cfg:
            assert isinstance(cfg, ConfigBase)
            self.set_options(cfg)
            self.set_options(kwargs)
        self.init()
        self.check_required()
    def set_options(self, cfg):
        if not isinstance(cfg,dict):
            cfg=cfg.to_dict()
        for k, v in cfg.items():
            if k in self.__allowed_options__:
                if v is None:
                    v= getattr(self,k,None)
                setattr(self, k, v)
    def check_required(self):
        for key in self.__required__:
            if getattr(self, key, None) is None:
                raise Exception('Argument %s is required but not given.' % (key))

    def init(self):
        if self.WEIGHTS_PATH:
            self.model.load_state_dict(torch.load(self.WEIGHTS_PATH))
        if self.DEVICE:
            self.model.to(self.DEVICE)
        self.model.eval()

class Predictor(PredictorBase):
    def predict(self,img):
        pass
    def preprocess(self,img):
        pass
    def postprocess(self,res):
        pass





from detro import utils as ut
from detro.predict import ClassicalHeatmapDetectorBase, ClassicalCenterOffsetsDetector


class TestDetector(ClassicalCenterOffsetsDetector):
    @classmethod
    def make_params(cls, input_size, num_points, num_classes=None, classes=None,
                    heatmap_downsample=4, score_thresh=0.1):
        params = ClassicalCenterOffsetsDetector.make_params(input_size=input_size, num_points=num_points,
                                                            num_classes=num_classes, classes=classes,
                                                            max_output_num=1,
                                                            heatmap_downsample=heatmap_downsample,
                                                            score_thresh=score_thresh)
        return params
    def get_middle_results(self):
        outputs=self.cache['outputs']
        center_heatmap_img=self.heatmap2image(outputs['center_heatmap'][0,0])
        corner_heatmap_img=self.heatmap2image(outputs['corner_heatmap'][0,0])
        return dict(
            center_heatmap_img=center_heatmap_img,corner_heatmap_img=corner_heatmap_img
        )
    def decode_outputs(self, outputs, down, *args, **kwargs):
        num_points = self.params.num_points
        heatmap = outputs['center_heatmap']
        offsets = outputs['offsets']
        return ut.decode_outputs_polygondet(heatmap, offsets, num_points=num_points, down=down,
                                            k=self.params.max_output_num, nms_radius=3)



import torch
import torch.nn as nn
import torch.nn.functional as F

import math
import numpy as np
# criterion = nn.BCEWithLogitsLoss(reduction='none')

def __focal(target, actual, alpha=1, gamma=2):
    focal = alpha * torch.pow(torch.abs(target - actual), gamma)
    return focal


def heatmap_loss(proto, coef, gt, ids):
    batch_size,num_object=ids.shape[:2]
    #print(proto.shape) # [16, 32, 128, 128]
    #print(coef.shape)  # [16, 32, 128, 128]
    #print(ids.shape) # [16, 6, 2]

    coef = F.softmax(coef, dim=1)
    ids=ids.long()
    arr=torch.zeros((coef.shape[0],ids.shape[1],coef.shape[1])).cuda()
    for batch_idx in range(coef.shape[0]):
        arr[batch_idx]=coef[batch_idx,:,ids[batch_idx,:,0],ids[batch_idx,:,1]].t()
    
    output = proto.view(batch_size, 1, 32, 128, 128) * arr.view(batch_size, num_object, 32, 1, 1)
    output=output.sum(dim=2)
    pred = torch.clamp(torch.sigmoid(output), min=1e-6, max=1-1e-6)

    pos_inds = gt.eq(1).float()
    neg_inds = gt.lt(1).float()

    neg_weights = torch.pow(1 - gt, 4)

    pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
    neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

    num_pos  = pos_inds.float().sum()
    pos_loss = pos_loss.sum()
    neg_loss = neg_loss.sum()

    if num_pos == 0:
        loss = -neg_loss
    else:
        loss = -(pos_loss + neg_loss) / num_pos
    return loss


def patch_loss(pred, gt):
    pred = torch.clamp(torch.sigmoid(pred), min=1e-6, max=1-1e-6)

    pos_inds = gt.eq(1).float()
    neg_inds = gt.lt(1).float()

    neg_weights = torch.pow(1 - gt, 4)

    pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
    neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

    num_pos  = pos_inds.float().sum()
    pos_loss = pos_loss.sum()
    neg_loss = neg_loss.sum()

    if num_pos == 0:
        loss = -neg_loss
    else:
        loss = -(pos_loss + neg_loss) / num_pos
    return loss


def center_loss(pred, gt):
    pred = torch.clamp(torch.sigmoid(pred), min=1e-6, max=1-1e-6)

    pos_inds = gt.eq(1).float()
    neg_inds = gt.lt(1).float()

    neg_weights = torch.pow(1 - gt, 4)

    pos_loss = torch.log(pred) * torch.pow(1 - pred, 2) * pos_inds
    neg_loss = torch.log(1 - pred) * torch.pow(pred, 2) * neg_weights * neg_inds

    num_pos  = pos_inds.float().sum()
    pos_loss = pos_loss.sum()
    neg_loss = neg_loss.sum()

    if num_pos == 0:
        loss = -neg_loss
    else:
        loss = -(pos_loss + neg_loss) / num_pos
    return loss

def l1_loss(pred,label):
    loss = F.l1_loss(pred, label, size_average=False) / (len(pred) + 1e-4)
    return loss

def wh_loss(pred_wh, label_wh, mask):
    '''
    preds (B x 4 x h x w)
    gt_regr (B x 4 x h x w)
    mask (B x 1 x h x w)
    '''
    loss = F.l1_loss(pred_wh * mask, label_wh, size_average=False) / (mask.sum() + 1e-4)

    # print(loss)

    # print(pred_wh.size())
    # print(label_wh.size())
    # loss = (torch.abs(pred_wh * mask - label_wh).sum(axis=(1, 2, 3)) / (mask.sum(axis=(1, 2, 3)) + 1e-4)).sum()
    
    return loss#  / len(pred_wh)
def distance_loss(pred_wh, label_wh, mask):
    '''
    preds (B x 4 x h x w)
    gt_regr (B x 4 x h x w)
    mask (B x 1 x h x w)
    '''
    loss = F.l1_loss(pred_wh * mask, label_wh, size_average=False) / (mask.sum() + 1e-4)

    # print(loss)

    # print(pred_wh.size())
    # print(label_wh.size())
    # loss = (torch.abs(pred_wh * mask - label_wh).sum(axis=(1, 2, 3)) / (mask.sum(axis=(1, 2, 3)) + 1e-4)).sum()

    return loss#  / len(pred_wh)


# def wh_loss(pred_lurd, label_lurd, mask):
#     '''
#     preds (B x 4 x h x w)
#     gt_regr (B x 4 x h x w)
#     mask (B x 1 x h x w)
#     '''
#     pred_coor = decode(pred_lurd)
#     label_coor = decode(label_lurd)
#     giou_loss = GIOU_loss(pred_coor, label_coor, mask)
#     loss = (giou_loss.sum(dim=(1,2,3)) / (mask.sum(dim=(1, 2, 3)) + 1e-4)).mean()
#     return loss




def GIOU(boxes1, boxes2):
    """
    input_format: (xmin, ymin, xmax, ymax)
    """
    boxes1 = torch.cat([torch.min(boxes1[..., :2], boxes1[..., 2:]),
                        torch.max(boxes1[..., :2], boxes1[..., 2:])], dim=-1)
    boxes2 = torch.cat([torch.min(boxes2[..., :2], boxes2[..., 2:]),
                        torch.max(boxes2[..., :2], boxes2[..., 2:])], dim=-1)

    boxes1_area = (boxes1[..., 2] - boxes1[..., 0]) * (boxes1[..., 3] - boxes1[..., 1])
    boxes2_area = (boxes2[..., 2] - boxes2[..., 0]) * (boxes2[..., 3] - boxes2[..., 1])
    # 计算出boxes1与boxes1相交部分的左上角坐标、右下角坐标
    intersection_left_up = torch.max(boxes1[..., :2], boxes2[..., :2])
    intersection_right_down = torch.min(boxes1[..., 2:], boxes2[..., 2:])

    # 因为两个boxes没有交集时，(right_down - left_up) < 0，所以maximum可以保证当两个boxes没有交集时，它们之间的iou为0
    intersection = torch.max(intersection_right_down - intersection_left_up, torch.zeros_like(intersection_right_down))
    inter_area = intersection[..., 0] * intersection[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    IOU = 1.0 * inter_area / (union_area + 1e-5)

    enclose_left_up = torch.min(boxes1[..., :2], boxes2[..., :2])
    enclose_right_down = torch.max(boxes1[..., 2:], boxes2[..., 2:])
    enclose = torch.max(enclose_right_down - enclose_left_up, torch.zeros_like(enclose_left_up))
    enclose_area = enclose[..., 0] * enclose[..., 1]
    GIOU = IOU - 1.0 * (enclose_area - union_area) / (enclose_area + 1e-5)

    return GIOU


def GIOU_loss(pred_coor, label_coor, respond_bbox):
    giou = GIOU(pred_coor, label_coor).unsqueeze(-1)
    bbox_wh = label_coor[..., 2:] - label_coor[..., :2]
    bbox_loss_scale = 2.0 - 1.0 * bbox_wh[..., 0:1] * bbox_wh[..., 1:2] / (512 ** 2)
    GIOU_loss = respond_bbox * bbox_loss_scale * (1.0 - giou)
    return GIOU_loss


def bbox_ciou(boxes1, boxes2):
    '''
    计算ciou = iou - p2/c2 - av
    pred_xywh
    label_xywh
    举例时假设pred_xywh和label_xywh的shape都是(1, 4)
    '''

    # 变成左上角坐标、右下角坐标
    boxes1_x0y0x1y1 = boxes1
    boxes2_x0y0x1y1 = boxes2
    '''
    逐个位置比较boxes1_x0y0x1y1[..., :2]和boxes1_x0y0x1y1[..., 2:]，即逐个位置比较[x0, y0]和[x1, y1]，小的留下。
    比如留下了[x0, y0]
    这一步是为了避免一开始w h 是负数，导致x0y0成了右下角坐标，x1y1成了左上角坐标。
    '''
    boxes1_x0y0x1y1 = torch.cat((torch.min(boxes1_x0y0x1y1[..., :2], boxes1_x0y0x1y1[..., 2:]),
                             torch.max(boxes1_x0y0x1y1[..., :2], boxes1_x0y0x1y1[..., 2:])), dim=-1)
    boxes2_x0y0x1y1 = torch.cat((torch.min(boxes2_x0y0x1y1[..., :2], boxes2_x0y0x1y1[..., 2:]),
                             torch.max(boxes2_x0y0x1y1[..., :2], boxes2_x0y0x1y1[..., 2:])), dim=-1)

    # 两个矩形的面积
    boxes1_area = (boxes1_x0y0x1y1[..., 2] - boxes1_x0y0x1y1[..., 0]) * (
                boxes1_x0y0x1y1[..., 3] - boxes1_x0y0x1y1[..., 1])
    boxes2_area = (boxes2_x0y0x1y1[..., 2] - boxes2_x0y0x1y1[..., 0]) * (
                boxes2_x0y0x1y1[..., 3] - boxes2_x0y0x1y1[..., 1])

    # 相交矩形的左上角坐标、右下角坐标，shape 都是 (8, 13, 13, 3, 2)
    left_up = torch.max(boxes1_x0y0x1y1[..., :2], boxes2_x0y0x1y1[..., :2])
    right_down = torch.min(boxes1_x0y0x1y1[..., 2:], boxes2_x0y0x1y1[..., 2:])

    # 相交矩形的面积inter_area。iou
    inter_section = right_down - left_up
    inter_section = torch.where(inter_section < 0.0, inter_section*0, inter_section)
    inter_area = inter_section[..., 0] * inter_section[..., 1]
    union_area = boxes1_area + boxes2_area - inter_area
    iou = inter_area / union_area

    # 包围矩形的左上角坐标、右下角坐标，shape 都是 (8, 13, 13, 3, 2)
    enclose_left_up = torch.min(boxes1_x0y0x1y1[..., :2], boxes2_x0y0x1y1[..., :2])
    enclose_right_down = torch.max(boxes1_x0y0x1y1[..., 2:], boxes2_x0y0x1y1[..., 2:])

    # 包围矩形的对角线的平方
    enclose_wh = enclose_right_down - enclose_left_up
    enclose_c2 = torch.pow(enclose_wh[..., 0], 2) + torch.pow(enclose_wh[..., 1], 2)

    # 两矩形中心点距离的平方
    p2 = torch.pow(boxes1[..., 0] - boxes2[..., 0], 2) + torch.pow(boxes1[..., 1] - boxes2[..., 1], 2)

    # 增加av。分母boxes2[..., 3]可能为0，所以加上除0保护防止nan。
    atan1 = torch.atan(boxes1[..., 2] / boxes1[..., 3])
    temp_a = torch.where(boxes2[..., 3] > 0.0, boxes2[..., 3], boxes2[..., 3] + 1.0)
    atan2 = torch.atan(boxes2[..., 2] / temp_a)
    v = 4.0 * torch.pow(atan1 - atan2, 2) / (math.pi ** 2)
    a = v / (1 - iou + v)

    ciou = iou - 1.0 * p2 / enclose_c2 - 1.0 * a * v
    return ciou


def decode(output, stride=4):
    # output: [batchsize, 4, height, width]
    # left, up, right, down ===>>> x1, y1, x2, y2
    bz=output.shape[0]
    gridsize=output.shape[-1]

    output=output.permute(0, 2, 3, 1)
    output=output.view(bz, gridsize, gridsize, 4)
    leftup,rightdown=torch.split(output, [2,2], dim=3)

    shiftx=torch.arange(0,gridsize,dtype=torch.float32)
    shifty=torch.arange(0,gridsize,dtype=torch.float32)
    shifty,shiftx=torch.meshgrid([shiftx, shifty])
    shiftx=shiftx.unsqueeze(0).repeat(bz, 1,1)
    shifty=shifty.unsqueeze(0).repeat(bz, 1,1)

    xy_grid=torch.stack([shiftx,shifty], dim=3).cuda()

    x1y1=(xy_grid+0.5)*stride - leftup
    x2y2=(xy_grid+0.5)*stride + rightdown

    coors=torch.cat((x1y1,x2y2),dim=3)
    return coors





import torch.nn.functional as F
import torch
import numpy as np
from .polygon_utils import cartesian2polar, polar2cartesian


def heatmap_nms(heat, kernel=3):
    hmax = F.max_pool2d(heat, kernel, stride=1, padding=(kernel - 1) // 2)
    keep = (hmax == heat).float()
    return heat * keep


def get_reshaped_index(index, shape, coords=[]):
    dim = shape[-1]
    coord = index % dim
    index = index // dim
    coords = [coord] + coords
    shape = shape[:-1]
    if not len(shape):
        return coords
    else:
        return get_reshaped_index(index, shape, coords)


def heatmap_topk_with_distance(heatmap, distance, k):
    shape = heatmap.shape
    num_classes = shape[0]
    flat = heatmap.view((num_classes, -1))
    values, indexes = torch.topk(flat, k=k, dim=1)
    num_maps = distance.shape[0]
    dists = torch.transpose(distance, 0, 1)
    dists = torch.transpose(dists, 1, 2)
    dists = dists.reshape((-1, num_maps))
    dists = dists[indexes]
    indexes = indexes.numpy()
    indexes = [[get_reshaped_index(index, shape[1:]) for index in cls_indexes] for cls_indexes in indexes]
    indexes = np.array(indexes)
    values = values.numpy()
    dists = dists.numpy()
    return indexes, values, dists

def heatmap_topk_with_feature(heatmap, feature, k):
    shape = heatmap.shape
    num_classes = shape[0]
    flat = heatmap.view((num_classes, -1))
    values, indexes = torch.topk(flat, k=k, dim=1)
    num_maps = feature.shape[0]
    dists = torch.transpose(feature, 0, 1)
    dists = torch.transpose(dists, 1, 2)
    dists = dists.reshape((-1, num_maps))
    dists = dists[indexes]
    indexes = indexes.numpy()
    indexes = [[get_reshaped_index(index, shape[1:]) for index in cls_indexes] for cls_indexes in indexes]
    indexes = np.array(indexes)
    values = values.numpy()
    dists = dists.numpy()
    return indexes, values, dists


def heatmap_topk(heatmap, k):
    shape = heatmap.shape
    num_classes = shape[0]
    flat = heatmap.view((num_classes, -1))
    values, indexes = torch.topk(flat, k=k, dim=1)
    indexes = indexes.numpy()
    indexes = [[get_reshaped_index(index, shape[1:]) for index in cls_indexes] for cls_indexes in indexes]
    indexes = np.array(indexes)
    values = values.numpy()
    return indexes, values


def decode_outputs_polygondet(heatmap, distance, down, k=100, num_points=4, nms_radius=3):
    heatmap = heatmap.detach().cpu()
    distance = distance.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, nms_radius)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    for i in range(batch_size):
        indexes, scores, dists = heatmap_topk_with_distance(heatmap[i], distance[i], k=k)
        coords = []
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        indexes *= down
        # dists *= down

        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_points * 2 + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            cls_coords[:, 0:num_points * 2] = (
                    indexes_cls.reshape((-1, 1, 2)) + dists[cls_idx].reshape((-1, num_points, 2))).reshape(
                (-1, num_points * 2))
            cls_coords[:, num_points * 2] = scores[cls_idx]
            cls_coords[:, num_points * 2 + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)


def decode_outputs_circledet(heatmap, distance, down, k=100, nms_radius=3):
    heatmap = heatmap.detach().cpu()
    distance = distance.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, nms_radius)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    num_params=3
    for i in range(batch_size):
        indexes, scores, dists = heatmap_topk_with_feature(heatmap[i], distance[i], k=k)
        coords = []
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        indexes *= down
        # dists *= down
        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_params + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            cls_coords[:, 0:num_params] = np.c_[indexes_cls,dists[cls_idx]]
            cls_coords[:, num_params] = scores[cls_idx]
            cls_coords[:, num_params + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)


def decode_outputs_polygondet_polar(heatmap, distance, num_points, down, k=100):
    heatmap = heatmap.detach().cpu()
    distance = distance.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, 3)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    for i in range(batch_size):
        indexes, scores, dists = heatmap_topk_with_distance(heatmap[i], distance[i], k=k)
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        coords = []
        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_points * 2 + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            indexes_cls *= down
            # distance*=down
            cls_coords[:, 0:num_points * 2] = (
                    indexes_cls.reshape((-1, 1, 2)) + polar2cartesian(dists[cls_idx].reshape((-1, num_points, 2)),
                                                                      from_degree=True)
            ).reshape((-1, num_points * 2))
            cls_coords[:, num_points * 2] = scores[cls_idx]
            cls_coords[:, num_points * 2 + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)


def decode_outputs_fcosdet(heatmap, distance, down, k=100):
    num_points = 2
    heatmap = heatmap.detach().cpu()
    distance = distance.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, 3)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    for i in range(batch_size):
        indexes, scores, dists = heatmap_topk_with_distance(heatmap[i], distance[i], k=k)
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        indexes *= down
        # dists *= down
        coords = []
        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_points * 2 + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            cls_coords[:, 0:num_points * 2] = (
                    indexes_cls.reshape((-1, 1, 2)) + (dists[cls_idx] * np.array([-1, -1, 1, 1])).reshape(
                (-1, num_points, 2))
            ).reshape((-1, num_points * 2))
            cls_coords[:, num_points * 2] = scores[cls_idx]
            cls_coords[:, num_points * 2 + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)


def decode_outputs_centerdet(heatmap, distance, down, k=100):
    num_points = 2
    heatmap = heatmap.detach().cpu()
    distance = distance.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, 3)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    for i in range(batch_size):
        indexes, scores, dists = heatmap_topk_with_distance(heatmap[i], distance[i], k=k)
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        indexes *= down
        # dists *= down
        coords = []
        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_points * 2 + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            cls_coords[:, 0:num_points * 2] = np.concatenate(
                (indexes_cls - dists[cls_idx] / 2, indexes_cls + dists[cls_idx] / 2), axis=1)
            cls_coords[:, num_points * 2] = scores[cls_idx]
            cls_coords[:, num_points * 2 + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)


def decode_outputs_polygondet_single(heatmap, num_points, down, nms_radius=3):
    '''heatmap corner'''
    k = 1
    heatmap = heatmap.detach().cpu()
    heatmap = torch.sigmoid(heatmap)
    heatmap = heatmap_nms(heatmap, nms_radius)
    batch_size = heatmap.shape[0]
    num_classes = heatmap.shape[1]
    res = []
    for i in range(batch_size):
        indexes, scores = heatmap_topk(heatmap[i], k=num_points)
        indexes = np.array(indexes) + np.array([0.5, 0.5])
        indexes *= down
        coords = []
        for cls_idx in range(num_classes):
            cls_coords = np.zeros((k, num_points * 2 + 2))
            indexes_cls = indexes[cls_idx][:, ::-1]
            cls_coords[0, 0:num_points * 2] = indexes_cls.reshape((-1))
            cls_coords[0, num_points * 2] = scores[cls_idx].min()
            cls_coords[0, num_points * 2 + 1] = cls_idx
            coords.append(cls_coords)
        coords = np.concatenate(coords, axis=0)
        res.append(coords)
    return np.array(res)



import torch.nn as nn
import torch.nn.functional as F



class FPN(nn.Module):
    """
    Feature Pyramid Network.

    This is an implementation of - Feature Pyramid Networks for Object
    Detection (https://arxiv.org/abs/1612.03144)

    Args:
        in_channels (List[int]): Number of input channels per scale.
        out_channels (int): Number of output channels (used at each scale)
        num_outs (int): Number of output scales.
        start_level (int): Index of the start input backbone level used to
            build the feature pyramid. Default: 0.
        end_level (int): Index of the end input backbone level (exclusive) to
            build the feature pyramid. Default: -1, which means the last level.
        add_extra_convs (bool): Whether to add conv layers on top of the
            original feature maps. Default: False.
        extra_convs_on_inputs (bool): Whether to apply extra conv on
            the original feature from the backbone. Default: False.
        relu_before_extra_convs (bool): Whether to apply relu before the extra
            conv. Default: False.
        no_norm_on_lateral (bool): Whether to apply norm on lateral.
            Default: False.
        conv_cfg (dict): Config dict for convolution layer. Default: None.
        norm_cfg (dict): Config dict for normalization layer. Default: None.
        act_cfg (str): Config dict for activation layer in ConvModule.
            Default: None.

    Example:
        >>> import torch
        >>> in_channels = [2, 3, 5, 7]
        >>> scales = [340, 170, 84, 43]
        >>> inputs = [torch.rand(1, c, s, s)
        ...           for c, s in zip(in_channels, scales)]
        >>> self = FPN(in_channels, 11, len(in_channels)).eval()
        >>> outputs = self.forward(inputs)
        >>> for i in range(len(outputs)):
        ...     print(f'outputs[{i}].shape = {outputs[i].shape}')
        outputs[0].shape = torch.Size([1, 11, 340, 340])
        outputs[1].shape = torch.Size([1, 11, 170, 170])
        outputs[2].shape = torch.Size([1, 11, 84, 84])
        outputs[3].shape = torch.Size([1, 11, 43, 43])
    """

    def __init__(self,
                 in_channels,
                 out_channels,
                 num_outs,
                 start_level=0,
                 end_level=-1,
                 add_extra_convs=False,
                 extra_convs_on_inputs=True,
                 relu_before_extra_convs=False,
                 no_norm_on_lateral=False,
                 conv_cfg=None,
                 norm_cfg=None,
                 act_cfg=None):
        super(FPN, self).__init__()
        assert isinstance(in_channels, list)
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.num_ins = len(in_channels)
        self.num_outs = num_outs
        self.relu_before_extra_convs = relu_before_extra_convs
        self.no_norm_on_lateral = no_norm_on_lateral
        self.fp16_enabled = False

        if end_level == -1:
            self.backbone_end_level = self.num_ins
            assert num_outs >= self.num_ins - start_level
        else:
            # if end_level < inputs, no extra level is allowed
            self.backbone_end_level = end_level
            assert end_level <= len(in_channels)
            assert num_outs == end_level - start_level
        self.start_level = start_level
        self.end_level = end_level
        self.add_extra_convs = add_extra_convs
        self.extra_convs_on_inputs = extra_convs_on_inputs

        self.lateral_convs = nn.ModuleList()
        self.fpn_convs = nn.ModuleList()

        for i in range(self.start_level, self.backbone_end_level):
            l_conv = ConvModule(
                in_channels[i],
                out_channels,
                1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg if not self.no_norm_on_lateral else None,
                act_cfg=act_cfg,
                inplace=False)
            fpn_conv = ConvModule(
                out_channels,
                out_channels,
                3,
                padding=1,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                act_cfg=act_cfg,
                inplace=False)

            self.lateral_convs.append(l_conv)
            self.fpn_convs.append(fpn_conv)

        # add extra conv layers (e.g., RetinaNet)
        extra_levels = num_outs - self.backbone_end_level + self.start_level
        if add_extra_convs and extra_levels >= 1:
            for i in range(extra_levels):
                if i == 0 and self.extra_convs_on_inputs:
                    in_channels = self.in_channels[self.backbone_end_level - 1]
                else:
                    in_channels = out_channels
                extra_fpn_conv = ConvModule(
                    in_channels,
                    out_channels,
                    3,
                    stride=2,
                    padding=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                    inplace=False)
                self.fpn_convs.append(extra_fpn_conv)

    # default init_weights for conv(msra) and norm in ConvModule
    def init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                xavier_init(m, distribution='uniform')

    @auto_fp16()
    def forward(self, inputs):
        assert len(inputs) == len(self.in_channels)

        # build laterals
        laterals = [
            lateral_conv(inputs[i + self.start_level])
            for i, lateral_conv in enumerate(self.lateral_convs)
        ]

        # build top-down path
        used_backbone_levels = len(laterals)
        for i in range(used_backbone_levels - 1, 0, -1):
            prev_shape = laterals[i - 1].shape[2:]
            laterals[i - 1] += F.interpolate(
                laterals[i], size=prev_shape, mode='nearest')

        # build outputs
        # part 1: from original levels
        outs = [
            self.fpn_convs[i](laterals[i]) for i in range(used_backbone_levels)
        ]
        # part 2: add extra levels
        if self.num_outs > len(outs):
            # use max pool to get more levels on top of outputs
            # (e.g., Faster R-CNN, Mask R-CNN)
            if not self.add_extra_convs:
                for i in range(self.num_outs - used_backbone_levels):
                    outs.append(F.max_pool2d(outs[-1], 1, stride=2))
            # add conv layers on top of original feature maps (RetinaNet)
            else:
                if self.extra_convs_on_inputs:
                    orig = inputs[self.backbone_end_level - 1]
                    outs.append(self.fpn_convs[used_backbone_levels](orig))
                else:
                    outs.append(self.fpn_convs[used_backbone_levels](outs[-1]))
                for i in range(used_backbone_levels + 1, self.num_outs):
                    if self.relu_before_extra_convs:
                        outs.append(self.fpn_convs[i](F.relu(outs[-1])))
                    else:
                        outs.append(self.fpn_convs[i](outs[-1]))
        return tuple(outs)
